#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "earthworm.h"
#include "kom.h"
#include "time_ew.h"
#include "pdl2ew.h"
#include "strfuncs.h"

typedef enum _myarg_indexes {
   FIRST_ARG,
   DELETE_ARG = FIRST_ARG,
   ORIGIN_TIME_ARG,
   LATITUDE_ARG,
   LONGITUDE_ARG,
   DEPTH_ARG,
   MAGNITUDE_ARG,
   NUM_PHASES_ARG,
   AZIMUTH_GAP_ARG,
   MIN_DIST_ARG,
   HORZ_ERR_ARG,
   VERT_ERR_ARG,
   EVENTSOURCE_ARG,
   EVENTSOURCECODE_ARG,
   VERSION_ARG,
   MAGTYPE_ARG,
   NUM_ARGS
} myarg_indexes;

char *myargs[NUM_ARGS] = {
    "--delete",
    "--property-eventtime=",
    "--property-latitude=",
    "--property-longitude=",
    "--property-depth=",
    "--property-magnitude=",
    "--property-num-phases-used=",
    "--property-azimuthal-gap=",
    "--property-minimum-distance=",
    "--property-horizontal-error=",
    "--property-vertical-error=",
    "--property-eventsource=",
    "--property-eventsourcecode=",
    "--property-version=",
    "--property-magnitude-type="};

#define VERSION_STR "1.0.2 - 2020-09-17"

#ifdef _SOLARIS
#define read_line(LP, SZ, FP) (fgets(LP, SZ, FP) != NULL)
#else
#define read_line(LP, SZ, FP) (getline(&LP, (size_t *)&SZ, FP) != -1)
#endif

#define BUFFER_SIZE 1024

#define SUBJECT_LINE_PREFIX "Subject: "

#define PERCENT_VAR_TEXT "%%"
#define ID_VAR_TEXT "%ID"
#define LINK_VAR_TEXT "%LINK"

#define DEFAULT_DELETE_BODY_TEXT                                               \
   "This is an automated computer-generated message notifying users of the "   \
   "USGS' Product Distribution Layer system that a previously disseminated "   \
   "event with ID %ID has been deleted from the PDL system\nUSGS Event Link: " \
   "%LINK%ID\n"
#define DEFAULT_EVENT_LINK "https://earthquake.usgs.gov/earthquakes/eventpage/"
#define DEFAULT_SUBJECT_PREFIX "EWalert"

static int Debug; /* non-zero -> debug logging         */
static FILE *LogFile;
static char DestDirectory[NAM_LEN];
static char TmpDirectory[NAM_LEN];
static char PathToEventEmail[NAM_LEN];
static char ExcludeNetworks[NAM_LEN];
static char EventLink[TEXT_LEN];
static char DeleteBodyText[BUFFER_SIZE - 80]; // buffer size minus some
                                              // characters for echo and subject
static int DeleteBodyTextLen;
static char SubjectPrefix[TEXT_LEN];

static void logDate(void)
{
    time_t now;
    struct tm res;
    char   date[10];
    time( &now );
    gmtime_ew( &now, &res );
    /******************* Y2K *********************
     * Add 1900 to tm_year since it represents   *
     * number of years since 1900                *
     ******************* Y2K *********************/
    sprintf (date, "%4d%02d%02d", (res.tm_year + TM_YEAR_CORR),
         (res.tm_mon + 1), res.tm_mday);
    fprintf (LogFile, "%s_UTC_%02d:%02d:%02d ",
             date, res.tm_hour, res.tm_min, res.tm_sec );
}

static void logMessage(const char * format, ...)
{
  va_list args;
  logDate();
  va_start (args, format);
  vfprintf (LogFile, format, args);
  va_end (args);
}

/******************************************************************************
 *  config() processes command file(s) using kom.c functions;         *
 *                    exits if any errors are encountered.                    *
 ******************************************************************************/
#define ncommand 3
static void config(char *ProgName, char *configfile) {
   char init[ncommand]; /* init flags, one byte for each required command */
   int nmiss;           /* number of required commands that were missed   */
   char *com;
   char *str;
   char processor[20];
   int nfiles;
   int success;
   int i;

   Debug = 0;
   LogFile = NULL;
   *DestDirectory = '\0';
   *TmpDirectory = '\0';
   *PathToEventEmail = '\0';
   *ExcludeNetworks = '\0';
   strncpy(EventLink, DEFAULT_EVENT_LINK, sizeof(EventLink));
   strncpy(DeleteBodyText, DEFAULT_DELETE_BODY_TEXT, sizeof(DeleteBodyText));
   DeleteBodyTextLen = strlen(DeleteBodyText);
   strncpy(SubjectPrefix, DEFAULT_SUBJECT_PREFIX, sizeof(SubjectPrefix));
   /* Set to zero one init flag for each required command
    *****************************************************/
   for (i = 0; i < ncommand; i++)
      init[i] = 0;

   /* Open the main configuration file
    **********************************/
   nfiles = k_open(configfile);
   if (nfiles == 0) {
      fprintf(
          stderr, "%s: Error opening command file <%s>; exiting!\n", ProgName,
          configfile);
      exit(-1);
   }
   
   /* Process all command files
    ***************************/
   while (nfiles > 0) /* While there are command files open */
   {
      while (k_rd()) /* Read next line from active file  */
      {
         com = k_str(); /* Get the first token from line */

         /* Ignore blank lines & comments
          *******************************/
         if (!com)
            continue;
         if (com[0] == '#')
            continue;

         /* Open a nested configuration file
          **********************************/
         if (com[0] == '@') {
            success = nfiles + 1;
            nfiles = k_open(&com[1]);
            if (nfiles != success) {
               fprintf(
                   stderr, "%s: Error opening command file <%s>; exiting!\n",
                   ProgName, &com[1]);
               exit(-1);
            }
            continue;
         }
         strcpy(processor, "config");

         /* Process anything else as a command
          ************************************/
         /*0*/ if (k_its("LogFile")) {
            str = k_str();
            if (str) {
               LogFile = fopen(str, "a");
               if (LogFile == NULL) {
                  fprintf(
                      stderr, "%s: Error opening log file <%s>; exiting!\n",
                      ProgName, str);
                  exit(-1);
               }
               init[0] = 1;
               if (Debug)
                  printf("LogFile (%s)\n", str);
            }
         }
         /*1*/ else if (k_its("DestDirectory")) {
            str = k_str();
            if (str) {
               strncpy(DestDirectory, str, sizeof(DestDirectory));
               // make sure DestDirectory exists
               if( CreateDir( DestDirectory ) != EW_SUCCESS ) {
                  logit( "e", "%s: trouble creating DestDirectory: %s\n",
                         ProgName, DestDirectory );
                  exit(-1);
               }
               init[1] = 1;
               if (Debug)
                  printf("DestDirectory (%s)\n", DestDirectory);
            }
         }
         /*2*/ else if (k_its("PathToEventEmail")) {
            str = k_str();
            if (str) {
               strncpy(PathToEventEmail, str, sizeof(PathToEventEmail));
               init[2] = 1;
               if (Debug)
                  printf("PathToEventEmail (%s)\n", PathToEventEmail);
            }
         }
         /*opt*/ else if (k_its("TmpDirectory")) {
            str = k_str();
            if (str) {
               strncpy(TmpDirectory, str, sizeof(TmpDirectory));
            }
         }
         /*opt*/ else if (k_its("Debug")) {
            Debug = k_int();
            if (Debug)
               printf("Debug (%d)\n", Debug);
         }
         /*opt*/ else if (k_its("ExcludeNetworks")) {
            str = k_str();
            if (str) {
               strncpy(ExcludeNetworks, str, sizeof(ExcludeNetworks));
               if (Debug)
                  printf("ExcludeNetworks (%s)\n", ExcludeNetworks);
            }
         }
         /*opt*/ else if (k_its("EventLink")) {
            str = k_str();
            if (str) {
               strncpy(EventLink, str, sizeof(EventLink));
               if (Debug)
                  printf("EventLink (%s)\n", EventLink);
            }
         }
         /*opt*/ else if (k_its("DeleteBodyFile")) {
            str = k_str();
            if (str) {
               FILE *fp = fopen(str, "r");
               if (fp == NULL) {
                  fprintf(
                      stderr,
                      "%s: Error opening DeleteBodyFile <%s>; exiting!\n",
                      ProgName, str);
                  exit(-1);
               }
               memset(DeleteBodyText, 0, sizeof(DeleteBodyText));
               fread(DeleteBodyText, 1, sizeof(DeleteBodyText) - 1, fp);
               DeleteBodyTextLen = strlen(DeleteBodyText);
               if (DeleteBodyTextLen == 0) {
                  if (ferror(fp)) {
                     fprintf(
                         stderr,
                         "%s: Error reading DeleteBodyFile <%s>: %s; "
                         "exiting!\n",
                         ProgName, str, strerror(errno));
                  } else {
                     fprintf(
                         stderr,
                         "%s: Error reading DeleteBodyFile <%s>; exiting!\n",
                         ProgName, str);
                  }
                  exit(-1);
               }
               fclose(fp);
               if (Debug)
                  printf("DeleteBodyFile (%s)\n", str);
            }
         }
         /*opt*/ else if (k_its("SubjectPrefix")) {
            str = k_str();
            if (str) {
               strncpy(SubjectPrefix, str, sizeof(SubjectPrefix));
               if (Debug)
                  printf("SubjectPrefix (%s)\n", SubjectPrefix);
            }
         }

         /* Unknown command
          *****************/
         else {
            fprintf(
                stderr, "%s: <%s> Unknown command in <%s>.\n", ProgName, com,
                configfile);
            continue;
         }

         /* See if there were any errors processing the command
          *****************************************************/
         if (k_err()) {
            fprintf(
                stderr, "%s: Bad <%s> command for %s() in <%s>; exiting!\n",
                ProgName, com, processor, configfile);
            exit(-1);
         }
      }
      nfiles = k_close();
   }

   /* After all files are closed, check init flags for missed commands
    ******************************************************************/
   nmiss = 0;
   for (i = 0; i < ncommand; i++)
      if (!init[i])
         nmiss++;
   if (nmiss) {
      fprintf(stderr, "%s: ERROR, no ", ProgName);
      if (!init[0])
         fprintf(stderr, "<LogFile> ");
      if (!init[1])
         fprintf(stderr, "<DestDirectory> ");
      if (!init[2])
         fprintf(stderr, "<PathToEventEmail> ");
      fprintf(stderr, "command(s) in <%s>; exiting!\n", configfile);
      exit(-1);
   }
    
   if ( !*TmpDirectory ) { // if temporary directory was not specified
      strncpy(TmpDirectory, DestDirectory, sizeof(TmpDirectory));
      strcat(TmpDirectory, "/tmp");
   }
   // make sure TmpDirectory exists
   if( CreateDir( TmpDirectory ) != EW_SUCCESS ) {
       logit( "e", "%s: trouble creating TmpDirectory: %s\n",
             ProgName, TmpDirectory );
       exit(-1);
   }
   if (Debug)
       printf("TmpDirectory (%s)\n", TmpDirectory);
}

/*
 *
 */
int main(int argc, char **argv) {
   int istart, iend, i, j;
   myarg_indexes myarg_index;
   double mag = 0;
   double val, aval, deg, min;
   char codaSet = 0;
   ARCDATA msg;
   int reqd = 2;
   int isDelete = 0;
   char eventEmailFileName[250] = {'\0'};
   FILE *eventEmailFile;
   char *linePtr = NULL;
   int linePtrSize = 0;
   char buffer[BUFFER_SIZE];
   char eventID[sizeof(msg.eventID) + 1] = {'\0'};
   char *str;
   char c;
   int string_len;

   if (argc < 2) {
      fprintf(
          stderr, "Usage: pdl2ark <configfile> <data-arg1> <data-arg2> ...\n");
      fprintf(stderr, "Version: %s\n", VERSION_STR);
      exit(1);
   }

   ARCDATA_INIT(msg);
   msg.version[0] = '1'; // default to version 1
   /* Read the configuration file(s)
   ********************************/
   config(argv[0], argv[1]);

   iend = -1;
   if ( Debug == 99 ) {
       logDate();
       fprintf(LogFile, "args:\n");
       for (istart = 2; istart < argc; istart++) {
           str = argv[istart];
           fprintf(LogFile, "%02d %s\n", istart, str);
       }
   }
   for (istart = 2; istart < argc; istart++) {
      str = argv[istart];
      iend = istart;
      // if string is quoted at start
      if (*str == '\"') {
         linePtr = str;
         linePtrSize = strlen(linePtr);
         str++; // skip the quote
         // while line is not quoted at end
         while (linePtr[linePtrSize - 1] != '\"') {
            if (++iend >= argc)
               break;
            linePtr = argv[iend];
            linePtrSize = strlen(linePtr);
         }
      }
      // find the myarg idx
      *buffer = '\0';
      for (myarg_index = FIRST_ARG; myarg_index < NUM_ARGS; myarg_index++) {
         string_len = strlen(myargs[myarg_index]);
         // if myargs match was found
         if (strncmp(myargs[myarg_index], str, string_len) == 0) {
            str += string_len; // skip the myargs text
            strcat(buffer, str);
            // copy the arg values
            for (j = istart + 1; j <= iend; j++) {
               str = argv[j];
               strcat(buffer, " ");
               strcat(buffer, str);
            }
            string_len = strlen(buffer);
            // if value ends with quote character
            if (string_len > 0 && buffer[string_len - 1] == '"') {
               // remove the quote character
               buffer[--string_len] = '\0';
            }
            // trim space if needed
            str = strchr(buffer, ' ');
            if (str)
               *str = '\0';
            break;
         }
      }

      string_len = 0; // default to no characters to copy
      str = NULL;
      switch (myarg_index) {
      case DELETE_ARG:
         isDelete = 1;
         break;
      case ORIGIN_TIME_ARG:
         string_len = strnlen(buffer, sizeof(msg.originTime));
         for (str = buffer, i = 0; i < string_len; str++) {
            c = *str;
            if (c == '\0')
               break;
            if (isdigit(c)) {
               msg.originTime[i++] = c;
            }
         }
         break;
      case LATITUDE_ARG:
         val = strtod(buffer, NULL);
         aval = fabs(val);
         deg = floor(aval);
         min = 60.0 * (aval - deg);
         sprintf(
             buffer, "%2.0lf%c%04.0lf", deg, val >= 0 ? 'N' : 'S', min * 100);
         string_len = strnlen(buffer, sizeof(msg.lat));
         str = msg.lat;
         break;
      case LONGITUDE_ARG:
         val = strtod(buffer, NULL);
         aval = fabs(val);
         deg = floor(aval);
         min = 60.0 * (aval - deg);
         sprintf(
             buffer, "%3.0lf%c%04.0lf", deg, val >= 0 ? 'E' : 'W', min * 100);
         string_len = strnlen(buffer, sizeof(msg.lon));
         str = msg.lon;
         break;
      case DEPTH_ARG:
         val = strtod(buffer, NULL);
         sprintf(buffer, "%5.0lf", val * 100);
         string_len = strnlen(buffer, sizeof(msg.z));
         str = msg.z;
         break;
      case MAGNITUDE_ARG:
         mag = strtod(buffer, NULL);
         break;
      case NUM_PHASES_ARG:
         val = strtod(buffer, NULL);
         sprintf(buffer, "%3.0lf", val);
         string_len = strnlen(buffer, sizeof(msg.nph));
         str = msg.nph;
         break;
      case AZIMUTH_GAP_ARG:
         val = strtod(buffer, NULL);
         sprintf(buffer, "%3.0lf", val);
         string_len = strnlen(buffer, sizeof(msg.gap));
         str = msg.gap;
         break;
      case MIN_DIST_ARG:
         val = strtod(buffer, NULL);
         sprintf(buffer, "%3.0lf", val);
         string_len = strnlen(buffer, sizeof(msg.dmin));
         str = msg.dmin;
         break;
      case HORZ_ERR_ARG:
         val = strtod(buffer, NULL);
         sprintf(buffer, "%4.0lf", val * 100);
         string_len = strnlen(buffer, sizeof(msg.erh));
         str = msg.erh;
         break;
      case VERT_ERR_ARG:
         val = strtod(buffer, NULL);
         sprintf(buffer, "%4.0lf", val * 100);
         string_len = strnlen(buffer, sizeof(msg.erz));
         str = msg.erz;
         break;
      case EVENTSOURCE_ARG:
         string_len = strnlen(buffer, 2);
         if (string_len != 2) {
             logMessage("Invalid source code %s\n",
                     buffer);
             exit(1);
         }
         memcpy(eventID, buffer, string_len);
         if ( (c = *ExcludeNetworks) ) {
             int match = 0;
             i = 0;
             j = 0;
             do {
                 if (c == ',') {
                     if (match > 0) {
                         break;
                     }
                     match = 0;
                     i = 0;
                     continue;
                 }
                 if (match < 0) { // if difference was found
                     continue; // continue until next network
                 }
                 if (c != eventID[i++]) {
                     match = -1;
                     i = 0;
                     continue;
                 }
                 match = 1;
             } while ( (c = ExcludeNetworks[++j]) );
             if (match > 0) {
                if (Debug)
                   puts("Excluded network");
                 return 0;
             }
         }
         reqd--;
         break;
      case EVENTSOURCECODE_ARG:
         strncpy(eventID + 2, buffer, sizeof(eventID) - 2);
         reqd--;
         break;
      case VERSION_ARG:
         string_len = strnlen(buffer, sizeof(msg.version));
         memcpy(msg.version, buffer, string_len);
         string_len = strnlen(buffer, sizeof(msg.eventVersion));
         str = msg.eventVersion;
         break;
      case MAGTYPE_ARG:
         codaSet = 1;
         break;
      default:
         break;
      }
      if (string_len > 0 && str) {
         memcpy(str, buffer, string_len);
      }
      istart = iend; // continue after end
   }
    
   if (reqd != 0) {
       logMessage("Origin but no EventID (%s)\n", eventID);
       exit(0);
   }

   string_len = strnlen(eventID, sizeof(msg.eventID));
   memcpy(msg.eventID, eventID, string_len);

   if (codaSet) {
      sprintf(buffer, "%3.0lf", mag * 100);
      string_len = strnlen(buffer, sizeof(msg.Md));
      memcpy(msg.Md, buffer, string_len);
      string_len = strnlen(buffer, sizeof(msg.Mpref));
      memcpy(msg.Mpref, buffer, string_len);
   }
   if ( !isDelete ) { // not delete
      char tmppath[800];
      char path[800];
      FILE *f2;
      sprintf(tmppath, "%s/%s.ark", TmpDirectory, eventID);
      *path = '\0';
      f2 = fopen(tmppath, "wb");
      if (f2 == NULL) {
         logMessage("Could not open file (%s): %s\n", tmppath, strerror(errno));
      } else {
         if (Debug) {
            puts(path);
            puts((char *)&msg);
         }
         if (fwrite((void *)&msg, 1, ARCFILE_SIZE, f2) != ARCFILE_SIZE) {
              logMessage("Could not write temporary file (%s): %s\n", tmppath, strerror(errno));
         } else {
             sprintf(path, "%s/%s.ark", DestDirectory, eventID);
         }
         if (fclose(f2) != 0) {
              logMessage("Could not close temporary file (%s): %s\n", tmppath, strerror(errno));
         } else if ( *path ) {
             if (rename_ew(tmppath, path) != 0) {
                 logMessage("Could not rename temporary file (%s->%s): %s\n", tmppath, path, strerror(errno));
             }
         } else {
             if (remove(tmppath) != 0) {
                 logMessage("Could not remove temporary file (%s): %s\n", tmppath, strerror(errno));
             }
         }
      }
   } else { // delete
      // check directory for file with email addresses this event has been sent
      // to
      sprintf(eventEmailFileName, "%s/%s", PathToEventEmail, eventID);
      if (access(eventEmailFileName, F_OK) == 0) {
         char firstLine[128];
         // here we need to send out emails to notify of a deletion
         // open the event-email mapping file
         if (!(eventEmailFile = fopen(eventEmailFileName, "r"))) {
            logMessage("Error opening event email file %s\n",
                eventEmailFileName);
            exit(1);
         }
         memset(firstLine, 0, sizeof(firstLine));
         linePtr = firstLine;
         linePtrSize = sizeof(firstLine);
         // if first line could not be read
         if (!read_line(linePtr, linePtrSize, eventEmailFile) ||
             *strstrip(firstLine) == '\0') {
            logMessage("Error reading event email file %s\n",
                    eventEmailFileName);
            exit(1);
         }
         memset(buffer, 0, BUFFER_SIZE);
         string_len = strlen(buffer);
         str = buffer + string_len;
         snprintf(str, BUFFER_SIZE - string_len, "echo \"");
         for (i = 0; i < DeleteBodyTextLen;) {
            string_len = strlen(buffer);
            str = buffer + string_len;
            c = DeleteBodyText[i];
            if (c != '%') {
               i++;
               *str++ = c;
               *str = '\0';
            } else if (
                strncmp(
                    DeleteBodyText + i, PERCENT_VAR_TEXT,
                    strlen(PERCENT_VAR_TEXT)) == 0) {
               i += strlen(PERCENT_VAR_TEXT);
               *str++ = c;
               *str = '\0';
            } else if (
                strncmp(DeleteBodyText + i, ID_VAR_TEXT, strlen(ID_VAR_TEXT)) ==
                0) {
               i += strlen(ID_VAR_TEXT);
               strncpy(str, eventID, BUFFER_SIZE - string_len);
            } else if (
                strncmp(
                    DeleteBodyText + i, LINK_VAR_TEXT, strlen(LINK_VAR_TEXT)) ==
                0) {
               i += strlen(LINK_VAR_TEXT);
               strncpy(str, EventLink, BUFFER_SIZE - string_len);
            } else {
               logMessage("Invalid DeleteBodyText:\n%s\n", DeleteBodyText);
               exit(1);
            }
         }
         string_len = strlen(buffer);
         str = buffer + string_len;
         snprintf(
	         str, BUFFER_SIZE - string_len, "\" | mailx -s \"");
         string_len = strlen(buffer);
         str = buffer + string_len;
         // if first line contains the subject
         if (strncmp(firstLine, SUBJECT_LINE_PREFIX, strlen(SUBJECT_LINE_PREFIX)) == 0) {
            snprintf(
               str, BUFFER_SIZE - string_len, "DELETED %s\" ",
               firstLine + strlen(SUBJECT_LINE_PREFIX));
            *firstLine = '\0'; // done with subject text
         } else {
            snprintf(
               str, BUFFER_SIZE - string_len, "DELETED %s - EW Event NEIC ID: %s\" ",
               SubjectPrefix, eventID);
         }
         // save the email address at the end of the system command
         linePtrSize = BUFFER_SIZE - strlen(buffer);
         linePtr = buffer + strlen(buffer);
         if (*firstLine != '\0') { // if first line was not subject
            // copy first line to buffer
            strcpy(linePtr, firstLine);
         } else {
            *linePtr = '\0';
         }
         for (i = 0;; i++, *linePtr = '\0') {
            if (*linePtr == '\0') { // if next line is needed
               // read the next line
               if (!read_line(linePtr, linePtrSize, eventEmailFile)) {
                  break; // line not available
               }
               strstrip(linePtr);
            }
            // linePtr contains an email address
            if (*linePtr == '\0') { // if empty email address
               logMessage("Empty email address (%d)\n", i);
               break;
            }
            if (Debug == 999) {
               printf("%d (%s)\n", i, buffer);
               continue;
            }
            if (Debug)
               logMessage("mailing delete message to: %s\n", linePtr);
            if (system(buffer)) {
               logMessage("mailx failure. EventID: %s\n Command: %s\n",
                   eventID, buffer);
               exit(1);
            }
         }
         fclose(eventEmailFile);
         if (Debug != 999)
            remove(eventEmailFileName);
      } else if (Debug) {
         logMessage("Received delete message for EventID %s, but no record exists of "
             "this having been emailed\n",
             eventID);
      }
   }
   fclose(LogFile);
   return 0;
}
