/******************************************************
  gps.c

  Functions for reading time from a Trimble TSIP GPS.
  These functions are Linux and Windows compatible.
 ******************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <sys/timeb.h>

#include <earthworm.h>
#include <time_ew.h>
#include <swap.h>
#include "adsendxs.h"

#define TIMESTRSIZE 26

/* Functions prototypes
   ********************/
void ReportErrorToStatmgr( int errNum, char *errmsg );
int  PurgeComInput( void );
void GetTimeFromPTP( uint8_t ptp[], time_t *tgps );
int  WriteGpsPacket( uint8_t packet[], int nchar );
int  ReadGpsPacket( uint8_t packet[] );
void DecodeGpsPacket( uint8_t packet[], int packetsize, uint8_t *packet_id, uint8_t dsb[], int *ndsb );
void RemoveDleBytes( uint8_t dsb[], int *ndsb );

/* Declared in getconfig.h
   ***********************/
extern int Debug;             // If > 0, log GPS flags


/***************************************************************
  DecodeGpsPacket

  Extract packet id and data string bytes from the GPS packet.
  dsb  = Pointer to data string bytes
  ndsb = Number of data string bytes
 ***************************************************************/

void DecodeGpsPacket( uint8_t packet[], int packetsize, uint8_t *packet_id,
                      uint8_t dsb[], int *ndsb )
{
   *packet_id = packet[1];
   *ndsb = packetsize - 4;
   memcpy( dsb, packet+2, *ndsb );
   return;
}


/********************************************************************
  RemoveDleBytes

  Remove any extra (stuffed) DLE bytes from the packet data string.
 ********************************************************************/

void RemoveDleBytes( uint8_t dsb[], int *ndsb )
{
   int newNdsb = *ndsb;
   int i;

   for ( i = 1; i < newNdsb; i++ )
   {
      if ( (dsb[i-1] == DLE) && (dsb[i] == DLE) )
      {
         newNdsb--;
         memmove( &dsb[i], &dsb[i+1], (newNdsb - i) );
      }
   }
   *ndsb = newNdsb;
   return;
}


/*********************************************************
  DisableAutoPackets

  Disable automatic output packets from the GPS.
  By default, the primary and supplemental timing
  packets are sent after every one-second pulse.
 *********************************************************/

int DisableAutoPackets( void )
{
   uint8_t gpsPacket[] = {DLE, 0x8E, 0xA5, 0, 0, 0, 0, DLE, ETX};
   int     packetsize;
   uint8_t inPacket[MAXGPSPACKETSIZE];

/* Purge com port input buffer
   ***************************/
   if ( PurgeComInput() < 0 )
   {
      logit( "et", "PurgeComInput error.\n" );
      return -1;
   }

/* Disable packet broadcasts
   *************************/
   if ( WriteGpsPacket( gpsPacket, 9 ) < 0 )
   {
      logit( "et", "WriteGpsPacket error in DisableAutoPackets.\n" );
      return -1;
   }

/* Read the response packet
   ************************/
   if ( (packetsize = ReadGpsPacket(inPacket)) < 0 )
   {
      logit( "et", "ReadGpsPacket error in DisableAutoPackets.\n" );
      return -1;
   }
   if ( packetsize != 9 )
   {
      logit( "et", "Error.  DisableAutoPacket packet length: %d\n", packetsize );
      logit( "et", "Packet length should be 9.\n" );
      return -1;
   }
   return 0;
}


/*************************************************************
  SetUtcMode

  Put the GPS into "UTC mode", instead of "GPS mode".
  This will cause times to be reported in UTC, and the
  one-PPS pulse will be referenced to UTC.
  Save the timing settings to non-volatile memory.
 *************************************************************/

int SetUtcMode( void )
{
   uint8_t modePacket[] = {DLE, 0x8E, 0xA2, 0x03, DLE, ETX};
   uint8_t savePacket[] = {DLE, 0x8E, 0x4C, 0x06, DLE, ETX};

   if ( WriteGpsPacket( modePacket, 6 ) < 0 )
   {
      logit( "et", "WriteGpsPacket error setting UTC mode.\n" );
      return -1;
   }

   if ( WriteGpsPacket( savePacket, 6 ) < 0 )
   {
      logit( "et", "WriteGpsPacket error saving timing settings to NVM.\n" );
      return -1;
   }
   return 0;
}


/*********************************************************
  GetPpsOutputStatus

  Query the PPS output status.
 *********************************************************/

int GetPpsOutputStatus( void )
{
   static  uint8_t requestPacket[] = {DLE, 0x8E, 0x4E, DLE, ETX};
   int     packetsize;
   uint8_t inPacket[MAXGPSPACKETSIZE];
   uint8_t packet_id;                  // GPS packet id
   uint8_t dsb[10];
   int     ndsb;
   const   int nptp = 2;               // Num bytes in PPS status packet

/* Purge com port input buffer
   ***************************/
   if ( PurgeComInput() < 0 )
   {
      logit( "et", "PurgeComInput error.\n" );
      return -1;
   }

/* Request PPS characteristics
   ***************************/
   if ( WriteGpsPacket( requestPacket, 5 ) < 0 )
   {
      logit( "et", "WriteGpsPacket error.\n" );
      return -1;
   }

/* Read the response packet
   ************************/
   if ( (packetsize = ReadGpsPacket(inPacket)) < 0 )
   {
      logit( "et", "ReadGpsPacket error in GetPpsOutputStatus.\n" );
      return -1;
   }
   if ( packetsize < 4 )
   {
      logit( "et", "Error.  GetPpsOutputStatus packet length: %d\n", packetsize );
      return -1;
   }

/* Extract packet id and data-string bytes from the incoming packet.
   dsb  = Data string bytes from GPS packet
   ndsb = Number of data string bytes
   ****************************************************************/
   DecodeGpsPacket( inPacket, packetsize, &packet_id, dsb, &ndsb );

   if ( packet_id != 0x8F )           // Wrong packet id
   {
      logit( "et", "Error. packet_id = %02X  (should be 0x8E)\n", packet_id );
      return -1;
   }

/* Remove any stuffed DLE bytes from the data string bytes.
   Then, check dsp bytes for sanity.
   *******************************************************/
   RemoveDleBytes( dsb, &ndsb );

   if ( ndsb != nptp )        // Wrong number of bytes in packet data string
   {
      int i;
      logit( "et", "Error. Number of data bytes in packet= %d (should be %d)\n", ndsb, nptp );
      for ( i = 0; i < ndsb; i++ )    // For debugging
         logit( "e", " %02X", dsb[i] );
      logit( "e", "\n" );
      return -1;
   }

   if ( dsb[0] != requestPacket[2] )  // Wrong packet subcode
   {
      logit( "et", "Error. Packet subcode for PTP = %02X  (should be %02X)\n",
             dsb[0], requestPacket[2] );
      return -1;
   }
   if ( dsb[1] & 0x02 )
      logit( "et", "The 1-PPS pulse is always on.\n" );
   else
      logit( "et", "1-PPS status byte: 0x%02x\n", dsb[1] );
   return 0;
}


/********************************************************************
  GetPrimaryTimingPacket

  Request that the primary timing packet be returned immediately.
  Returns 0 if successful.
 ********************************************************************/

int GetPrimaryTimingPacket( uint8_t dsb[] )
{
   static  uint8_t requestPacket[] = {DLE, 0x8E, 0xAB, 0x00, DLE, ETX};
   int     packetsize;
   uint8_t inPacket[MAXGPSPACKETSIZE];
   uint8_t packet_id;                // GPS packet id
   int     ndsb;
   const   int nptp = 17;            // Num bytes in primary timing packet

/* Purge com port input buffer
   ***************************/
   if ( PurgeComInput() < 0 )
   {
      logit( "et", "PurgeComInput error.\n" );
      return -1;
   }

/* Request primary timing packet
   *****************************/
   if ( WriteGpsPacket( requestPacket, 6 ) < 0 )
   {
      logit( "et", "WriteGpsPacket error.\n" );
      return -1;
   }

/* Read the primary timing packet
   ******************************/
   if ( (packetsize = ReadGpsPacket(inPacket)) < 0 )
   {
      logit( "et", "ReadGpsPacket error in GetPrimaryTimingPacket.\n" );
      return -1;
   }
   if ( packetsize < 4 )
   {
      logit( "et", "Error.  GetPrimaryTimingPacket packet length: %d\n", packetsize );
      return -1;
   }

/* Extract packet id and data-string bytes from the incoming packet.
   dsb  = Data string bytes from GPS packet
   ndsb = Number of data string bytes
   ****************************************************************/
   DecodeGpsPacket( inPacket, packetsize, &packet_id, dsb, &ndsb );

   if ( packet_id != 0x8F )           // Wrong packet id
   {
      logit( "et", "Error. packet_id = %02X  (should be 0x8F)\n", packet_id );
      return -1;
   }

/* Remove any stuffed DLE bytes from the data string bytes.
   Then, check dsp bytes for sanity.
   *******************************************************/
   RemoveDleBytes( dsb, &ndsb );

   if ( ndsb != nptp )        // Wrong number of bytes in packet data string
   {
      int i;
      logit( "et", "Error. Number of data bytes in packet= %d (should be %d)\n", ndsb, nptp );
      for ( i = 0; i < ndsb; i++ )    // For debugging
         logit( "e", " %02X", dsb[i] );
      logit( "e", "\n" );
      return -1;
   }

   if ( dsb[0] != requestPacket[2] )  // Wrong packet subcode
   {
      logit( "et", "Error. Packet subcode for PTP = %02X  (should be %02X)\n",
             dsb[0], requestPacket[2] );
      return -1;
   }
   return 0;
}


/**********************************************************************
  GetSupplementalTimingPacket

  Request that the supplemental timing packet be returned immediately.
  Returns 0 if successful.
 **********************************************************************/

int GetSupplementalTimingPacket( uint8_t dsb[] )
{
   static  uint8_t requestPacket[] = {DLE, 0x8E, 0xAC, 0x00, DLE, ETX};
   int     packetsize;
   uint8_t inPacket[MAXGPSPACKETSIZE];
   uint8_t packet_id;                // GPS packet id
   int     ndsb;
   const   int nstp = 68;            // Num bytes in supplemental timing packet

/* Purge com port input buffer
   ***************************/
   if ( PurgeComInput() < 0 )
   {
      logit( "et", "PurgeComInput error.\n" );
      return -1;
   }

/* Request supplemental timing packet
   **********************************/
   if ( WriteGpsPacket( requestPacket, 6 ) < 0 )
   {
      logit( "et", "WriteGpsPacket error.\n" );
      return -1;
   }

/* Read the supplemental timing packet
   ***********************************/
   if ( (packetsize = ReadGpsPacket(inPacket)) < 0 )
   {
      logit( "et", "ReadGpsPacket error in GetSupplementalTimingPacket.\n" );
      return -1;
   }
   if ( packetsize < 4 )
   {
      logit( "et", "Error.  GetSupplementalTimingPacket packet length: %d\n", packetsize );
      return -1;
   }

/* Extract packet id and data-string bytes from the incoming packet.
   dsb  = Data string bytes from GPS packet
   ndsb = Number of data string bytes
   ****************************************************************/
   DecodeGpsPacket( inPacket, packetsize, &packet_id, dsb, &ndsb );

   if ( packet_id != 0x8F )           // Wrong packet id
   {
      logit( "et", "Error. packet_id = %02X  (should be 0x8F)\n", packet_id );
      return -1;
   }

/* Remove any stuffed DLE bytes from the data string bytes.
   Then, check dsp bytes for sanity.
   *******************************************************/
   RemoveDleBytes( dsb, &ndsb );

   if ( ndsb != nstp )        // Wrong number of bytes in packet data string
   {
      logit( "et", "Error. Number of data bytes in packet= %d (should be %d)\n", ndsb, nstp );
      return -1;
   }

   if ( dsb[0] != requestPacket[2] )  // Wrong packet subcode
   {
      logit( "et", "Error. Packet subcode for STP = %02X  (should be %02X)\n",
             dsb[0], requestPacket[2] );
      return -1;
   }
   return 0;
}


/******************************************************************************************
  GetGpsTime

  Returns antennaOpen. Possible values are ANTENNA_OPEN_UNKNOWN, ANTENNA_OPEN_FALSE,
                       or ANTENNA_OPEN_TRUE.

  Returns CRITICAL_ERROR if no GPS time is available, and we want to exit the program.
  Returns GPS_TIME_UNAVAILABLE if no GPS time is available.
  Returns UNLOCKED_GPS_TIME_AVAILABLE if time is available, but no satellite lock.
  Returns LOCKED_GPS_TIME_AVAILABLE if time is available, and clock is locked to satellites.
 ******************************************************************************************/

int GetGpsTime( time_t *tgps, int *antennaOpen )
{
   uint8_t  ptp[MAXGPSPACKETSIZE];    // Primary timing packet
   uint8_t  stp[MAXGPSPACKETSIZE];    // Supplemental timing packet
   uint8_t  timingFlags;              // From PTP
   uint8_t  discipliningMode;         // From STP
   uint16_t minorAlarms;              // From STP
   uint8_t  gpsDecodingStatus;        // From STP
   int      ptpError, stpError;
   int      i;
   char     errMsg[ERRMSGSIZE];

/* Trimble constants
   *****************/
   const uint8_t no_utc_info              = 0x08;   // PTP Timing Flag
   const uint8_t locked_to_gps            = 0;      // STP Disciplining Mode
   const uint8_t power_up                 = 1;      // STP Disciplining Mode
   const uint8_t auto_holdover            = 2;      // STP Disciplining Mode
   const uint8_t recovering_discipline    = 4;      // STP Disciplining Mode
// const uint8_t doing_fixes              = 0x00;   // STP GPS Decoding Status (unused)
   const uint8_t no_usable_satellites     = 0x08;   // STP GPS Decoding Status
   const uint8_t only_1_usable_satellite  = 0x09;   // STP GPS Decoding Status
   const uint8_t only_2_usable_satellites = 0x0A;   // STP GPS Decoding Status
   const uint8_t only_3_usable_satellites = 0x0B;   // STP GPS Decoding Status

   *antennaOpen = ANTENNA_OPEN_UNKNOWN;

/* Read Primary Timing Packet (8F-AB) from GPS.
   A Critical Error will cause the program to exit.
   ***********************************************/
   ptpError = GetPrimaryTimingPacket( ptp );
   if ( Debug == 2 )      // Log complete PTP data
   {
      logit( "e", "\nptp:" );
      for ( i = 0; i < 17; i++ )
         logit( "e", " %02X", ptp[i] );
      logit( "e", "\n" );
   }
   if ( ptpError )
   {
      logit( "et", "Error reading Primary Timing Packet from GPS.\n" );
      strncpy( errMsg, "Error reading Primary Timing Packet from GPS.\n", ERRMSGSIZE );
      ReportErrorToStatmgr( GPS_ERROR, errMsg );
      return CRITICAL_ERROR;
   }

   memcpy( &timingFlags, ptp+9, 1 );
   if ( Debug == 1 )           // Log timing flags
      logit( "e", "\n" );
   if ( Debug > 0 )
      logit( "e", "timingFlags: %02X\n", timingFlags );

/* Read Supplemental Timing Packet (8F-AC) from GPS.
   A Critical Error will cause the program to exit.
   ************************************************/
   stpError = GetSupplementalTimingPacket( stp );
   if ( Debug == 2 )      // Log complete STP data
   {
      logit( "e", "stp:" );
      for ( i = 0; i < 60; i++ )
         logit( "e", " %02X", stp[i] );
      logit( "e", "\n" );
   }
   if ( stpError )
   {
      logit( "et", "Error reading Supplemental Timing Packet from GPS.\n" );
      strncpy( errMsg, "Error reading Supplemental Timing Packet from GPS.\n", ERRMSGSIZE );
      ReportErrorToStatmgr( GPS_ERROR, errMsg );
      return CRITICAL_ERROR;
   }

   memcpy( &discipliningMode, stp+2, 1 );
   memcpy( &minorAlarms, stp+10, 2 );
   SwapShort( &minorAlarms );
   memcpy( &gpsDecodingStatus, stp+12, 1 );
   if ( Debug > 0 )
   {
      logit( "e", "discipliningMode:%02X", discipliningMode );
      logit( "e", "  minorAlarms:%04X", minorAlarms );
      logit( "e", "  gpsDecodingStatus:%02X\n", gpsDecodingStatus );
   }
   *antennaOpen = (minorAlarms & 0x0002) ? ANTENNA_OPEN_TRUE : ANTENNA_OPEN_FALSE;

/* If the "DAC is at rail", exit program
   *************************************/
   {
      uint16_t criticalAlarms;
      memcpy( &criticalAlarms, stp+8, 2 );
      SwapShort( &criticalAlarms );
      if ( criticalAlarms & 0x0010 )
      {
         logit( "et", "Error.  GPS DAC at rail.\n" );
         strncpy( errMsg, "Error.  GPS DAC at rail.\n", ERRMSGSIZE );
         ReportErrorToStatmgr( GPS_ERROR, errMsg );
         return CRITICAL_ERROR;
      }
   }

/* Log any changes in disciplining mode, which is an indication
   of GPS receiver health, satellite health, and antenna cable
   health.  In debug mode, log disciplining mode every second.
   ************************************************************/
   {
      static uint8_t discipliningModePrev = 5;   // Mode 5 not used by Trimble
      static int first = 1;                      // First time we ran this code

      if ( (Debug > 0) || (discipliningMode != discipliningModePrev) )
      {
         if ( discipliningMode == locked_to_gps )
            logit( "et", "GPS clock locked to satellites.\n" );
         else if ( discipliningMode == power_up )
            logit( "et", "GPS powering up.\n" );
         else if ( discipliningMode == auto_holdover )
            logit( "et", "GPS clock is undisciplined.\n" );
//       else if ( discipliningMode == recovering_discipline )
//          logit( "et", "GPS clock recovering discipline from satellites.\n" );
      }

/* If GPS clock loses or regains satellite synch, send an email alarm
   ******************************************************************/
      if ( (discipliningModePrev == locked_to_gps) &&
           (discipliningMode != locked_to_gps ) )
      {
         strncpy( errMsg, "Lost synch to GPS satellites.", ERRMSGSIZE );
         ReportErrorToStatmgr( GPS_ERROR, errMsg );
      }

      if ( !first &&
           (discipliningModePrev != locked_to_gps) &&
           (discipliningMode == locked_to_gps ) )
      {
         strncpy( errMsg, "Regained synch to GPS satellites.", ERRMSGSIZE );
         ReportErrorToStatmgr( GPS_ERROR, errMsg );
      }
      discipliningModePrev = discipliningMode;
      first = 0;
   }

/* Log if not enough usable satellites
   ***********************************/
   {
      static uint8_t gpsDecodingStatusPrev = 0xFF;    // 0xFF not used by Trimble

      if ( gpsDecodingStatus != gpsDecodingStatusPrev )
      {
         if ( gpsDecodingStatus == no_usable_satellites )
            logit( "et", "No usable satellites.\n" );
         else if ( gpsDecodingStatus == only_1_usable_satellite )
            logit( "et", "Only 1 usable satellite.\n" );
         else if ( gpsDecodingStatus == only_2_usable_satellites )
            logit( "et", "Only 2 usable satellites.\n" );
         else if ( gpsDecodingStatus == only_3_usable_satellites )
            logit( "et", "Only 3 usable satellites.\n" );
      }
      gpsDecodingStatusPrev = gpsDecodingStatus;
   }

/* See if accurate time is available, whether or not
   the GPS clock is locked to satellites.
   *************************************************/
   if ( discipliningMode & power_up )
   {
      logit( "et", "GPS powering up.\n" );
      return GPS_TIME_UNAVAILABLE;
   }
   if ( timingFlags & no_utc_info )
   {
      logit( "et", "No UTC info.\n" );
      return GPS_TIME_UNAVAILABLE;
   }
   if ( (discipliningMode != locked_to_gps) &&
        (discipliningMode != auto_holdover) &&
        (discipliningMode != recovering_discipline) )
      return GPS_TIME_UNAVAILABLE;

/* The Trimble clock time should now be ok.
   Get time from the PTP and return it to calling program.
   In debug mode, convert time to readable form using ctime.
   Warning: ctime_s may fail in 64-bit environments.
   ********************************************************/
   GetTimeFromPTP( ptp, tgps );
   if ( Debug > 0 )                               // Log GPS clock time
   {
      char timestr[TIMESTRSIZE];
//    if ( ctime_s( timestr, TIMESTRSIZE, tgps ) != 0 )   // Original call in Windows version
      if ( ctime_ew( tgps, timestr, TIMESTRSIZE ) == NULL )
      {
         logit( "et", "Error converting GPS time to string using ctime_ew\n" );
         return GPS_TIME_UNAVAILABLE;
      }
      timestr[24] = '\0';                         // Chop off trailing CR
      logit( "e", "GPS clock: %s\n", timestr+4 ); // Chop off day-of-week
   }

/* Is the GPS clock locked to satellites?
   **************************************/
   if ( discipliningMode == locked_to_gps )
      return LOCKED_GPS_TIME_AVAILABLE;
   else
      return UNLOCKED_GPS_TIME_AVAILABLE;
}


/*****************************************************
  GetTimeFromPTP

  Get UTC time from Primary Timing Packet.
  Convert to seconds since midnight Jan 1, 1970.
 *****************************************************/

void GetTimeFromPTP( uint8_t ptp[], time_t *tgps )
{
   struct tm stm;
   short  year;

   stm.tm_sec  = ptp[10];
   stm.tm_min  = ptp[11];
   stm.tm_hour = ptp[12];
   stm.tm_mday = ptp[13];
   stm.tm_mon  = ptp[14] - 1;
   memcpy( &year, ptp+15, 2 );
   SwapShort( &year );
   stm.tm_year = year - 1900;
   stm.tm_isdst = 0;             // Disable daylight savings mode
   *tgps = timegm_ew( &stm );    // Whole numbers of seconds
   return;
}
