
/*************************************************************************/
/*                                                                       */
/*	File: interpolate.c                                                   */
/*                                                                       */
/* Description: Calculate latency values for regularly-spaced time       */
/* intervals based on irregularly-spaced latency values read in from     */
/* a file.  Mean latency associated with a particular time value are     */
/* found via integration over an interval centered on that time value,   */
/* with reference latencies calculated by linear integration between     */
/* latency values in the file data.                                      */
/*                                                                       */
/*************************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define INTERVAL 1.0 // Interval (seconds) that we are integrating over

int debug = 1; // Toggle debug messages

double inputTime[1024]; // Array for storing input times
double inputLat[1024];  // Array for storing input latencies
double refTime; // Time (second) that we want to output latency value for
int nSamp;      // Utility counter

int getLatency();

int main(int argc, char **argv) {
  int nloop = 0; // Record number read in

  double t_in; //	Time value read from input record
  double l_in; // Latency value read from input record

  FILE *fin;                 // Input file
  fin = fopen(argv[1], "r"); // Open input file

  refTime = 0.;
  nSamp = 0;

  while (1) {
    ++nloop;

    if (debug) {
      printf("\nLoop %d:\n", nloop);
    }

    // Read one line from input file
    if (fscanf(fin, "%lf %lf", &t_in, &l_in) == EOF)
      break;

    if (debug) {
      printf("READ: %11.2f %5.2f\n", t_in, l_in);
      printf("refTime: %11.2f\n", refTime);
    }

    // Add read values to our arrays
    inputTime[nSamp] = t_in;
    inputLat[nSamp] = l_in;

    // Set initial value for refTime if it is not already set
    if (refTime == 0.) {
      refTime = (int)(t_in + 1.);
    }

    ++nSamp;

    getLatency(); // Calculate latency values

  } // end while (reading records)

  return 1;
} // end main

int getLatency() {
  int nShift;          // Count used to discard already-processed inputs
  int iBefore, iAfter; // Indices of inputs bracketing interval of interest
  int nOut;            // Utility counter
  int i;               // Utility counter
  double tStart, tEnd; // Start/end times for the interval being integrated over
  double time[1024];   // Times for point being integrated over
  double latency[1024]; // Latencies for point being integrated over
  double outputLatency; // Latency value calculated for a given refTime
  double frac;          // Fractional value used interpolation calculation

  while (1) {
    // Set interval start and end for the iteration based on refTime
    tStart = refTime - 0.5 * INTERVAL;
    tEnd = refTime + 0.5 * INTERVAL;

    if (debug) {
      printf("\n");
      printf("tStart: %11.2f\n", tStart);
      printf("tEnd: %11.2f\n", tEnd);
    }

    // Analyze arrays to determine whether we can output anything
    iBefore = 0;
    iAfter = 0;

    for (i = 0; i < nSamp;
         ++i) // For the records currently in our sample window
    {
      if (inputTime[i] <= tStart)
        iBefore = i; // Index of the last data value before (or at) tStart
    }

    // Shift values in the arrays (discard values that are not necessary
    // for future processing, to save space).  Any future values of 'tStart'
    // will be later than the current value of 'tStart', which is itself later
    // than the time pointed to by 'iBefore'.  Records preceding 'iBefore'
    // are therefore unnecessary.

    nShift = iBefore;

    for (i = nShift; i < nSamp; ++i) {
      inputTime[i - nShift] = inputTime[i];
      inputLat[i - nShift] = inputLat[i];
    }

    nSamp -= nShift; // Shift sampling window
    iBefore = 0;     // Reset index once shift has been performed

    for (i = 0; i < nSamp;
         ++i) // For the records currently in our sample window
    {
      if (inputTime[i] >= tEnd) {
        iAfter = i; // Index of the first data value after (or at) tEnd
        break;
      }
    }

    if (iBefore == iAfter && iAfter < (nSamp - 1))
      ++iAfter;

    if (debug) {
      printf("Held Input Times: ");
      for (i = 0; i < nSamp; ++i)
        printf("%11.2f ", inputTime[i]);
      printf("\n");

      printf("iBefore=%d, iAfter=%d\n", iBefore, iAfter);
      fflush(stdout);
    }

    if (inputTime[iAfter] < tEnd)
      break;

    // Calculate interpolation weighting fraction for beginning of time window
    frac = (tStart - inputTime[iBefore]) /
           (inputTime[iBefore + 1] - inputTime[iBefore]);

    if (frac < 0 || frac > 1.0) {
      printf("Error: Invalid fraction.\n");
      exit(0);
    }

    if (debug) {
      printf("frac(begin)=%f\n", frac);
    }

    nOut = 0;

    // Apply weighting and add interpolated latency to array used for
    // calculations
    time[nOut] = tStart;
    latency[nOut] =
        frac * inputLat[iBefore + 1] + (1. - frac) * inputLat[iBefore];
    ++nOut;

    for (i = iBefore + 1; i < iAfter; ++i) {
      time[nOut] = inputTime[i];
      latency[nOut] = inputLat[i];
      ++nOut;
    }

    // Re-calculate interpolation weighting fraction for end of time window
    frac = (tEnd - inputTime[iAfter - 1]) /
           (inputTime[iAfter] - inputTime[iAfter - 1]);

    if (frac < 0 || frac > 1.0) {
      printf("Error: Invalid fraction.\n");
      exit(0);
    }

    if (debug) {
      printf("frac(end)=%f\n", frac);
    }

    // Apply weighting and add interpolated latency to array used for
    // calculations
    time[nOut] = tEnd;
    latency[nOut] =
        frac * inputLat[iAfter] + (1. - frac) * inputLat[iAfter - 1];
    ++nOut;

    if (debug) {
      for (i = 0; i < nOut; ++i) {
        printf("time[%d]=%f, latency[%d]=%f\n", i, time[i], i, latency[i]);
      }
    }

    // Calculate mean latency as trapezoidal summation over time window
    outputLatency = 0;

    for (i = 1; i < nOut; ++i) {
      outputLatency +=
          0.5 * (latency[i] + latency[i - 1]) * (time[i] - time[i - 1]);
    }

    outputLatency /= INTERVAL;

    if (debug) {
      printf("iBefore=%d, iAfter=%d: refTime=%11.2f, outputLatency= %7.4f, "
             "nshift=%d\n",
             iBefore, iAfter, refTime, outputLatency, iBefore + 1);
    }

    // Output Record
    printf("Time: %11.2f, Latency: %7.4f\n", refTime, outputLatency);
    fflush(stdout);

    // Increment the reference time
    refTime += 1.;

    // If we do not have data to fill in intermediate times,
    // shift the reference time that we are looking for.
    if (refTime < inputTime[0]) {
      refTime = (int)(inputTime[0] + 1.);
    }
  } // end while (processing a time window)

  return 0;
}
