#include "scnl_table.h"
#include "earthworm.h"
#include <stdlib.h>
#include <string.h>

static const char TERM_CHARS[] = ". \r\n";
static const int NUM_TERM_CHARS = sizeof(TERM_CHARS);

/**
 * Deterrmines if the character is a terminating character for SCNL.
 * @param c the character
 * @return true if the character is a terminating character, false otherwise.
 */
int is_scnl_term(char c) {
  int i;
  for (i = 0; i < NUM_TERM_CHARS; i++) {
    if (c == TERM_CHARS[i]) {
      return i + 1;
    }
  }
  return 0;
}

/**
 * Add an entry to the SCNL table.
 * @param scnl_table the SCNL table.
 * @return a pointer to the empty new SCNL entry.
 */
SCNL_ENTRY *scnl_table_add(SCNL_TABLE *scnl_table) {
  SCNL_ENTRY *scnl_entry;
  if (scnl_table->nSCNL >= scnl_table->maxSCNL) {
    scnl_table_resize(scnl_table, scnl_table->maxSCNL + INCREMENT_SCNL);
  }
  scnl_entry = scnl_table->scnl_entries + scnl_table->nSCNL;
  scnl_table->nSCNL++;
  return scnl_entry;
}

/**
 * Copy the  string for a SCNL.
 * @param dest the destination string.
 * @param dest_size the size of the destination string.
 * @param src the source string.
 * @param len the lenth of the source string.
 */
void scnl_table_strcpy(char *dest, size_t dest_size, char *src, int len) {
  if (len >= dest_size) { // if len is too large
    len = dest_size - 1;  // truncate string
  }
  memcpy(dest, src, len);
  *(dest + len) = '\0'; // terminate the string
}

/**
 * Resize the SCNL table.
 * @param scnl_table the SCNL table.
 * @param maxSCNL the maximum number of SCNL entries.
 */
void scnl_table_resize(SCNL_TABLE *scnl_table, int maxSCNL) {
  size_t oldSize = scnl_table->maxSCNL * sizeof(SCNL_ENTRY);
  size_t newSize = maxSCNL * sizeof(SCNL_ENTRY);
  if (newSize <= oldSize) { // ensure new size is larger than old size
    return;
  }
  scnl_table->maxSCNL = maxSCNL;
  scnl_table->scnl_entries =
      (SCNL_ENTRY *)realloc(scnl_table->scnl_entries, newSize);
  if (scnl_table->scnl_entries == NULL) {
    logit("e",
          "scnl_table_add: Error allocating %zu bytes for SCNL entries; "
          "exiting!\n",
          newSize);
    exit(-1);
  } else {
    // clear new entries
    size_t diff = newSize - oldSize;
    void *pStart = ((char *)scnl_table->scnl_entries) + oldSize;
    memset(pStart, 0, diff);
  }
}

/**
 * Set the entry by parsing the specified string.
 * @param scnl_entry the SCNL entry to set.
 * @param str the string to parse the SCNL.
 * @return a pointer in the string after the SCNL.
 */
char *scnl_table_set_entry(SCNL_ENTRY *scnl_entry, char *str) {
  char *end = str;

  // get the station
  while (!is_scnl_term(*end)) {
    ++end;
  }
  scnl_table_strcpy(scnl_entry->sta, sizeof(scnl_entry->sta), str, end - str);

  str = end + 1;
  end = str;
  if (*str == '\0') {
    return str;
  }

  // get the channel
  while (!is_scnl_term(*end)) {
    ++end;
  }
  scnl_table_strcpy(scnl_entry->chan, sizeof(scnl_entry->chan), str, end - str);

  str = end + 1;
  end = str;
  if (*str == '\0') {
    return str;
  }

  // get the network
  while (!is_scnl_term(*end)) {
    ++end;
  }
  scnl_table_strcpy(scnl_entry->net, sizeof(scnl_entry->net), str, end - str);

  str = end + 1;
  end = str;
  if (*str == '\0') {
    return str;
  }

  // get the location
  while (!is_scnl_term(*end)) {
    ++end;
  }
  scnl_table_strcpy(scnl_entry->loc, sizeof(scnl_entry->loc), str, end - str);

  str = end + 1;

  return str;
}

/**
 * Determines the number of lines in the file and sets the table size
 * accordingly.
 * @param scnl_table the SCNL table.
 * @param fp the file pointer.
 */
void scnl_table_size(SCNL_TABLE *scnl_table, FILE *fp) {
  char c;
  int numChars = 0;
  int maxSCNL = 0;
  rewind(fp); // go to beginning
  while ((c = fgetc(fp)) != EOF) {
    if (c == '\n') {
      numChars = 0;
      maxSCNL++; // add an entry
    } else {
      numChars++;
    }
  }
  if (numChars > 0) { // if leftover characters
    maxSCNL++;        // add an entry
  }
  rewind(fp); // go to beginning
  scnl_table_resize(scnl_table, maxSCNL);
}
