/*
 * grid.c : grid manager for maximum likelyhood method
 */
#include <math.h>
#include <limits.h>
#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "chron3.h"
#include "kom.h"
#include "tlay.h"
#include "site.h"
#include "geo.h"
#include "grid.h"
#include "bind.h"
#include "earthworm_simple_funcs.h"

#define LAT(yy) ((yy) / grdFac.y + grdOrg.lat)
#define LON(xx) ((xx) / grdFac.x + grdOrg.lon)
#define X(LON) (grdFac.x * ((LON) - grdOrg.lon))
#define Y(LAT) (grdFac.y * ((LAT) - grdOrg.lat))
#define MAXDIS 100
#define Panic(x) (logit("e","Panic point %d in grid.c\n",(x)),exit(-1))

#define MAXLST 100

/* Function prototypes
   *******************/
int    GetInst( char *, unsigned char * );   /* getutil.c sys-independent  */
void   grid_init( void );
int    grid_cmpsrt( const void *, const void * );
float  pick_weight(int wt, unsigned char instid);
void   grid_free( void );
void   bndr_link( long, long );
void   bndr_quake2k( long );
double median( int, double * );
void pca2d(double *mdat, int nr, int nc, double *eval, double *evec);
int is_same_SCNL(int i1, int i2); 

/* Structure for sorting list of stacking picks on time
 ******************************************************/
typedef struct {
     double t;       /* arrival time of pick                     */
     short  ilist;   /* pre-sorted (original) index of this pick */
} SRT;               /* structure added 960408:ldd */

/* Structure for pick list (moved from grid_joint)
 ******************************************************/
typedef struct {                /* List of picks to stack, 0 is master  */
     long  id;       /* Pick id                              */
     float t;        /* Offset in tenths of seconds          */
     float x;        /* Station location (grid coord)        */
     float y;        /* Station location (grid coord)        */
     char  use;      /* 1 if pk should be stacked; 0 if not  */ /*960408:ldd*/
     char  horiz;    /* 1 horizontal component, 0 if vert */
     float wt;
} PLIST;

/* Structure for MLE 
 ******************************************************/
typedef struct {                /* List of maximum likelihood result    */
     double  xmax,ymax;         /* maxima of x, y                       */
     int     cx,cy,izmax;             /* maximum of z                   */
     double  mLikely;           /* maximum likelihood                   */
     double  fLikely;           /* maximum likelihood  of first layer   */
} MLres;



double student_hypb(double ti, double tobs, int dof);          /* SHEEN */
void qsort_id(double *a, int *id,int l,int r);                 /* SHEEN */
MLres calcMLE(int step, PLIST *list, int nlist);               /* SHEEN */
QUAKE testMLE(int qid, int npix, int *IPLIST);                 /* SHEEN */


/* Output control
   **************/
static int grid_debug = 0;
/* the following two settings are OPTIONAL rejection parameters triggered by nearest_quake_dist command */
static double nearest_quake_dist = 0.0;		/* epicentral km to nearest quake */
#define NEAREST_QUAKE_TIME 2.0
static double nearest_quake_time = NEAREST_QUAKE_TIME; 	/* seconds to nearest quake */
static int ignore_loc_code_in_same_station_decision = 0; /* in the is_same_SNL() function, if this is set to 1 ignore the location code as a tie breaker */

/* Stacking parameters
   *******************/
#define MAXWT 10
static int nWt = 0;
static struct {
        char          wt;       /* Weight code (0-4)                    */
        float         fwt;      /* Stack increment                      */
        unsigned char instid;   /* Installation id; added 20020523:ldd  */
} Wt[MAXWT];
unsigned char InstWild = 0;     /* wildcard for installation id         */ /*020523:ldd*/
char    InitInstWild   = 0;     /* initialization flag                  */ /*020523:ldd*/
double facCluster      = 4.0;   /* Pick clustering cutoff               */
double resGrid         = 1.0;   /* Tolerance for reassociating waifs    */
double rmsGrid         = 1.0;   /* RMS threshold for new quake          */
double GlitchNsec      = .035;  /* A glitch is defined as a group of at */ /*960408:ldd*/
int    GlitchMinPk     = 4;     /*    least GlitchMinPk picks within    */ /*960408:ldd*/
                                /*    GlitchNsec seconds                */ /*960408:ldd*/
double rMaxBind        = 200.0; /* Max. dist. for binding (200. km)     */
double tResBind        = 1.0;   /* Max. travel time residual 1.0 s      */
double minLikely       = 0.4;   /* Minimum likely value                 */
long   mJoint          = 20;    /* Maximum number of picks to stack     */
unsigned int maskGrid  = 0;     /* Initialization mask                  */
int    stack_horizontals = 0;   /* set to 1 to allow horizontal components in stack - paulf */
int    minPick          = 4;     /* minimum pick to associate an event   */
int    mPix;
int    nPix;

typedef struct {
    double lat;
    double lon;
    int    operating,si;
} OPSITE;
OPSITE *OPsite;
int check_pick_list(int is, char *stn, char *ntw, char *loc);
// ****************************/

static int nDist;               /* Dimensions of trav time template     */

int    nTrig           = 5;     /* minimum number of triggered stations for MAXEL */
float  fracDist        = 100.;  /* Event will not be associated if the fraction of */
                                /* triggered/all stations within fracDist distance  */
                                /* is smaller than nTrig */
double OPhour          = 1.;
int    nOPsite         = 0;
float  oPRate          = 0.5;
int    OPcheck         = 0;
float  oPDist          = 100.;

/* Grid definition parameters
 ****************************/
static int initGrid = 0;
static int nX, nY, nZ;          /* Dimensions of stacking buffer        */
float dTime;            /* Time granularity (seconds)                   */
float dSpace;           /* Space granularity (seconds)                  */
float dSpaceZ;          /* Vertical Space granularity (seconds)         */   // SHEEN
CART    grdTop;         /* Heigth, Width and Depth of grid              */
CART    grdStp;         /* Grid spacing, km                             */
CART    grdFac;         /* Geocentric to cartesian factor (km/degree)   */
GEO     grdOrg;         /* Origin = (0,0,0) in cartesian coordinates    */
GEO     grdMax;         /* Geocentric upper bound on grid               */

/* Travel-time table
 **********************/
/* Trav[ix][iz] : This array has  the travel time talbe from the grid
                to a station.  Index is calculated as
                iz*nDist + ix                                           */
double *dTrav;
double dTrstep;
int dTravMax; /* the maximum index in dTrav[]  */
int nTravGPD = 1; /* Number of Travel time grids per unit distance */

double *eDist;   /* temporay epicentral distance */
int    *eDid;    /* temporay epicentral distance */

/* Stacking Grid
 ***************/
/* Grid[ix][iy][iz]:  The space time coordinate of a box in the
                stack accumulator is taken to be the center of the box in
                space-time. Index calculated as iz + nZ*iy + nY*nZ*ix.
                The grid should be thought of as being
                draped over the pick initiating the association         */
double *dGrid;
int dGridMax;
static time_t sae=0;

/* Variable for saving last pick time
 **********************************/
double    *LPTime;




/* Functions to return index into 1-D array based on 3-D indices
 ***************************************************************/
#define IXGRID(IX, IY, IZ) ((IZ) + nZ * (IY) + nY * nZ * (IX))
int IXTRAV(double r, int iz)
{
        int ix = (int)(r*nTravGPD);

        if(ix < 0)      ix = -ix;
        return (iz*nDist*nTravGPD + ix);
}

/********************************************************************************
 * SHEEN                                                                        *
 * student_hypb(ti, tobs, dof) : Student's t-distribution                       *
 * ti: difference of predicted travel times                                     *
 * tobs: difference of observed times                                           *
 * dof: degree of freedom                                                       *
 ********************************************************************************/

double get_trtime(double r, int iz)
{
    int it, id;
    double p1,p2,rd, trt;

    it = IXTRAV( r, iz);
    if (it<0) it = 0;
    else if( it>=dTravMax ) it = dTravMax-1;
    p1 = dTrav[it];

    id = it%(nDist*nTravGPD);
    rd = dTrstep * id;

    it = IXTRAV( rd+dTrstep, iz);
    if (it<0) it = 0;
    else if( it>=dTravMax ) it = dTravMax-1;
    p2 = dTrav[it];

    trt = p1 + (p2-p1)/dTrstep*(r - rd);

    return trt;
}

double student_hypb(double ti, double tobs, int dof)
{
   double stdt;

   stdt = pow( 1. + pow(ti-tobs,2.)/dof, -0.5*(1.+dof) );
   return stdt;
}


double lhf_hypb(double xx, double yy, int iz, PLIST *list, int nlist)
{
    int j,it;
    double likely, tt, p1, p2;
    float wt1,wt2;

    likely = 1.;

    p1 = get_trtime( sqrt(pow(list[0].x-xx,2.)+pow(list[0].y-yy,2.)), iz);

    wt1 = list[0].wt;

    for(j=1;j<nlist;j++) {
        wt2 = list[j].wt;
        tt = list[0].t - list[j].t;

        p2 = get_trtime( sqrt(pow(list[j].x-xx,2.)+pow(list[j].y-yy,2.)), iz);

        likely *= wt1*wt2*student_hypb(p1-p2, tt, 1);
    }

    likely = pow(likely,1./(nlist-1.));
    return likely;
}

#define MAXPDF 500
double lhf_hypb2(double xx, double yy, int iz, PLIST *list, int nlist)
{
    int it,i1,i2,idx;
    int nblist;
    double likely, tt, p1, p2, lsum, wt1, wt2;
    double PDF[MAXPDF];

    if( nlist > 12 ) nblist = 12;       // SHEEN 2021.01.25
    else             nblist = nlist;       // SHEEN 2021.01.25

    for(i1=0;i1<nblist;i1++)
    {
        p1 = get_trtime( sqrt(pow(list[i1].x-xx,2.)+pow(list[i1].y-yy,2.)), iz);

        for(i2=i1+1;i2<nblist;i2++)
        {
            idx = nblist*i1 - i1*(i1+1)/2 + i2-i1-1;

            tt = list[i1].t - list[i2].t;

            p2 = get_trtime( sqrt(pow(list[i2].x-xx,2.)+pow(list[i2].y-yy,2.)), iz);

            PDF[idx] = student_hypb(p1-p2, tt, 1);
        }
    }

    lsum = 0.;
    for(i1=0; i1<nblist; i1++)
    {
//        wt1 = list[i1].wt;
        wt1 = 1.;
//        likely = 1.;
        likely = wt1;
        for(i2=0; i2<nblist; i2++)
        {
            if(i1==i2) continue;

            if(i1 > i2) idx = nblist*i2 - i2*(i2+1)/2 + i1-i2-1;
            else        idx = nblist*i1 - i1*(i1+1)/2 + i2-i1-1;

//            wt2 = list[i2].wt;
            wt2 = 1.;
            likely *= wt2*PDF[idx];
        }
        lsum += likely;
    }
    likely = lsum;

    likely = pow(likely/nblist,1./(nblist-1.));

    return likely;
}


void hGsearch(int sx, int ex, int sy, int ey, int iz, int step, MLres *mres, PLIST *list, int nlist)
{
    int ix, iy, is;
    double xx, yy;
    double dstep;

    dstep = dSpace*step;

    xx = sx*dstep;
    for(ix=sx; ix<=ex; ix+=step,xx+=dstep) {
        yy = sy*dstep;
        for(iy=sy; iy<=ey; iy+=step,yy+=dstep) {
            is = IXGRID(ix, iy, iz);
            if ((is<0) || (is>=dGridMax)) Panic( 3 );

            dGrid[is] =  lhf_hypb2(xx, yy, iz, list, nlist);
            if( mres->mLikely < dGrid[is] )
            {
                mres->mLikely = dGrid[is];
                mres->xmax = xx;
                mres->ymax = yy;
                mres->izmax = iz;
                mres->cx = ix;
                mres->cy = iy;
            }
        }
    }
}

MLres calcMLE(int step, PLIST *list, int nlist)                 
{
    int sx, ex, sy, ey, iz, ssize;
    MLres mres;

    mres.mLikely = 0.;
    ssize = 5;

    hGsearch(0,nX-1,0,nY-1, 0, step, &mres, list ,nlist);
    mres.fLikely = mres.mLikely;

    sx = mres.cx - ssize;
    if( sx < 0 ) sx = 0;
    ex = mres.cx + ssize;
    if( ex > nX-1 ) ex = nX-1;
    sy = mres.cy - ssize;
    if( sy < 0 ) sy = 0;
    ey = mres.cy + ssize;
    if( ey > nY-1 ) ey = nY-1;

    for(iz=1; iz<nZ; iz+=step) 
    {
        hGsearch(sx,ex, sy,ey, iz, step, &mres, list ,nlist);

        sx = mres.cx - ssize;
        if( sx < 0 ) sx = 0;
        ex = mres.cx + ssize;
        if( ex > nX-1 ) ex = nX-1;
        sy = mres.cy - ssize;
        if( sy < 0 ) sy = 0;
        ey = mres.cy + ssize;
        if( ey > nY-1 ) ey = nY-1;
    }

    return mres;
}

/********************************************************************************
 * grid_init() : Initialize grid calculations                                   *
 *              (once per execution after commands are read)                    *
 ********************************************************************************/
void grid_init( void )
{
        double t, z;
        double r;
        double dtdr, dtdz;
        long ix, iz;
        long ixyz;
        long i;

     /* Initialize the Stacking Grid! */
        if(!initGrid) {
                if ( maskGrid != 15 ) {
                    logit( "e", "grid_init: Cannot initialize; "  );
                    if( !(maskGrid&1)  ) logit( "e", "<grdlat> "  );
                    if( !(maskGrid&2)  ) logit( "e", "<grdlon> "  );
                    if( !(maskGrid&4)  ) logit( "e", "<grdz> "    );
                    if( !(maskGrid&8)  ) logit( "e", "<dspace> "  );
                    logit( "e", "command(s) omitted; exiting!\n" );
                    exit( -1 );
                }
                grdFac.y = (float)(40000.0 / 360.0);
                grdFac.x = grdFac.y
                        * cos(0.5*(grdOrg.lat+grdMax.lat)*6.283185/360.0);
                if(grdFac.x < (float)0.0)
                        grdFac.x = -grdFac.x;
                grdFac.z = (float)1.0;
                initGrid = 1;
                grid_cart(&grdMax, &grdTop);
                nX = (int)(grdTop.x / dSpace + 0.5);
                nY = (int)(grdTop.y / dSpace + 0.5);
                nZ = (int)(grdTop.z / dSpaceZ + 1.1);
                if(nZ < 1) nZ = 1;
                logit( "", "grid_init: entire stack grid dimensions: "
                           "nx, ny, nz = %d %d %d\n", nX, nY, nZ); 

                logit( "", "grid_range: entire grid ranges: "
                           "%.2f %.2f  %.2f %.2f  %.2f %.2f\n", 
                           grdOrg.lat, grdMax.lat, grdOrg.lon, grdMax.lon, grdOrg.z, grdMax.z); 

                logit( "e", "grid_init: minimum of triggered station "
                            "minimum = %d  distance_threshold: %.2f\n", nTrig, fracDist );

                logit( "e", "grid_init: minimum pick to associate an event:  %d \n", minPick );
                dGridMax = nX * nY * nZ;
                dGrid = (double *)malloc( nX*nY*nZ*sizeof(double) );
                for(i=0; i<nX*nY*nZ; i++) dGrid[i] = 1.;

                eDist = (double *)malloc(nSite*sizeof(double) );
                eDid  = (int    *)malloc(nSite*sizeof(int)    );

                nDist  = (int) sqrt( grdTop.x*grdTop.x + grdTop.y*grdTop.y );
                logit( "", "ttTable_init: nDist = %d, nTravGPD: %d, nZ: %d\n", nDist, nTravGPD, nZ);
                dTravMax = nDist * nZ * nTravGPD;
                dTrav = (double *)malloc( dTravMax*sizeof(double) );

                for(ix=0; ix<nDist*nTravGPD; ix++) {
                    r = (double)ix/(double)nTravGPD;
                    for(iz=0; iz<nZ; iz++) {
                        z = grdOrg.z + (double)iz*dSpaceZ;
                        if( ix==0) logit("","Depth grids for binder_max: %.3f\n",z);
                        t = t_lay(r, z, &dtdr, &dtdz);
                        ixyz = IXTRAV( r , iz );
                        if ((ixyz<0) || (ixyz>=dTravMax)) Panic( 11 );
                        dTrav[ixyz] = t;
                    }
                }
                logit( "", "grid_init: stack grid initialized: %.1f km\n", r);

    /* Allocate pick list LPTime */
                LPTime = (double *)malloc(nSite*sizeof(double));
                for(i=0; i<nSite; i++)
                        LPTime[i] = (double)GSEC1970;

                // initialize the maximum number of site for further use
                OPsite = (OPSITE *)malloc(nSite*sizeof(OPSITE));
    }
}

/**********************************************************************
 * grid_com()  Process all recognized commands.                       *
 **********************************************************************/
int grid_com( void )
{
        dTrstep = (double)1./nTravGPD;

        if( !InitInstWild ) {
           if( GetInst( "INST_WILDCARD", &InstWild ) != 0 ) {
               logit("e","grid_com: INST_WILDCARD not valid; does earthworm_global.d"
                        " file exist in EW_PARAMS?; exiting!\n" );
               exit( 0 );
           }
           InitInstWild = 1;
        }

        if(k_its("sae")) {
           sae = k_int();
           return 1;
	}
        if(k_its("ignore_loc_code_in_same_station_decision")) {
           ignore_loc_code_in_same_station_decision = k_int();
           return 1;
	}
        if(k_its("grid_debug")) {
           grid_debug = k_int();
           return 1;
	}
        if(k_its("nearest_quake_dist")) {
           nearest_quake_dist = k_val();	/* required distance threshold */
           nearest_quake_time = k_val();	/* optional time setting */
	   if (k_err()) nearest_quake_time = NEAREST_QUAKE_TIME; /* set to default */
           return 1;
	}

        if(k_its("grdlat")) {
           grdOrg.lat = (float)k_val();
           grdMax.lat = (float)k_val();
           maskGrid |= 1;
           return 1;
        }

        if(k_its("grdlon")) {
           grdOrg.lon = (float)k_val();
           grdMax.lon = (float)k_val();
           maskGrid |= 2;
           return 1;
        }

        if(k_its("grdz")) {
           grdOrg.z = (float)k_val();
           grdMax.z = (float)k_val();
           maskGrid |= 4;
           return 1;
        }

        if(k_its("dspace")) {
           dSpace  = (float)k_val();
           dSpaceZ = (float)k_val();
           maskGrid |= 8;
           return 1;
        }
        if(k_its("stack_horizontals")) {
	   stack_horizontals = 1;
           return 1;
	}

        if(k_its("thresh_triggered")) {
           nTrig = k_int();
           fracDist = (float)k_val();                     /* SHEEN */
           return 1;
        }

        if(k_its("min_likely")) {                       /* SHEEN */
           minLikely = (float)k_val();
           return 1;
        }

        if(k_its("rmax_bind")) {                       /* SHEEN */
           rMaxBind = (float)k_val();
           return 1;
        }

        if(k_its("tres_bind")) {
           tResBind = (float)k_val();                 /* SHEEN: Travel time residual threshold for removing outliers */
           return 1;
        }

        if(k_its("njoint")) {
           mJoint = k_long();                       /* SHEEN */
           return 1;
        }

        if(k_its("min_pick")) {
           minPick = k_int();                        /* SHEEN */
           return 1;
        }

        if(k_its("nTravGPD")) {
           nTravGPD = k_int();                        /* SHEEN */
           dTrstep = (double)1./nTravGPD;
           return 1;
        }

        if(k_its("rmsgrid")) {
           rmsGrid = k_val();
           return 1;
        }

        if(k_its("grid_wt")) {
           if(nWt < MAXWT) {
              Wt[nWt].wt     = k_int() + '0';
              Wt[nWt].fwt  = 1. - (4.-k_val())*0.2;
              if( Wt[nWt].fwt  < 1.e-3 ) Wt[nWt].fwt  = 0.1;     /* SHEEN */
              Wt[nWt].instid = InstWild;
              nWt++;
              return 1;
           }
           logit( "e", "grid_com: Too many <grid_wt> commands! "
                       "MAXWT = %d; exiting!\n", MAXWT );
           exit( 0 );
        }


        if( k_its("InOperation") ) {         /* SHEEN */
                OPhour = (double)k_val();
                if( OPhour > 0. )
                    OPcheck = 1;
                return 1;
        }
        if(k_its("InOprRate")) {
           oPDist = (float)k_val();                 /* SHEEN: operation distance */
           oPRate = (float)k_val();                 /* SHEEN: operation rate */
           return 1;
        }

        if(k_its("grid_wt_instid")) {
           char *str;
           if( nWt < MAXWT ) {
              Wt[nWt].wt  = k_int() + '0';
              Wt[nWt].fwt  = (float)k_val();
              str = k_str();
              if( str ) {
                 if( GetInst( str, &Wt[nWt].instid ) != 0 ) {
                    logit("e","grid_com: invalid installation id: %s in"
                              "<grid_wt_instid> command; exiting!\n", str );
                    exit( 0 );
                 }
              } else {
                 logit("e","grid_com: no installation id in "
                          "<grid_wt_instid> command; exiting!\n" );
                 exit( 0 );
              } 
              logit("e","grid_wt_instid: %c %3.1f %s:%u\n",
                     Wt[nWt].wt, Wt[nWt].fwt, str, Wt[nWt].instid );
              nWt++;
              return 1;
           }
           logit( "e", "grid_com: Too many <grid_wt_instid> commands! "
                       "MAXWT = %d; exiting!\n", MAXWT );
           exit( 0 );
        }

        if( k_its("define_glitch") ) {   /*new command; 960408:ldd*/
           GlitchMinPk = k_int();
           GlitchNsec  = k_val();
           return 1;
        }

        return 0;
}

/********************************************************************************
 * grid_cart(geo, cart)  Convert point from geocentric to cartesian coordinates *
 ********************************************************************************/
void grid_cart(GEO *geo, CART *cart)
{
        if(!initGrid)
                grid_init();
        cart->x = grdFac.x * (geo->lon - grdOrg.lon);
        cart->y = grdFac.y * (geo->lat - grdOrg.lat);
        cart->z = grdFac.z * (geo->z   - grdOrg.z);
}

/********************************************************************************
 * grid_geo(geo, cart) : Convert point from cartesian to geocentric coordinates *
 ********************************************************************************/
void grid_geo(GEO *geo, CART *cart)
{
        if(!initGrid)
                grid_init();
        geo->lon = cart->x / grdFac.x + grdOrg.lon;
        geo->lat = cart->y / grdFac.y + grdOrg.lat;
        geo->z   = cart->z + grdOrg.z;
}

/********************************************************************************
 * pick_weight()    Return stacking increment based on pick quality and            *
 *               installation id that produced the pick                         *
 ********************************************************************************/
float pick_weight(int wt, unsigned char instid)
{
        float weight;
        int i;

        weight = 1.;
        for(i=0; i<nWt; i++) {
           if(Wt[i].wt == wt) {
              if( Wt[i].instid == instid ) {   /* found rule for this installation */
                 weight = Wt[i].fwt;              
                 break;
              }
              if( Wt[i].instid == InstWild ) { /* wildcard: set increment but keep */
                 weight = Wt[i].fwt;              /* looking to see if there's a rule */
              }                                
           }
        }
        return weight;
}

/********************************************************************************
 * grid_joint(lpick) : 
 ********************************************************************************/
int grid_joint(long lpick)
{
        double  xx,yy,xmax,ymax,zmax,dthr,dtmp;
        int     ixpick,izmax;
        int     site;           /* Stacking phase, site index           */
        int     site_index;     /* Stackable pick, site index pointer   */
        int     ix1, ix2;       /* Stacking region, overlap of pTrav    */
        int     iy1, iy2;
        long    l1, l2;         /* Range of pick ids to stack           */
        long    l, l_internal;  /* Iterating pick index                 */
        int     ipick;          /* lpick % mPick                        */
                                /* list changed to global by SHEEN      */
        PLIST   list[MAXLST],tlist[MAXLST];
        SRT     srt[MAXLST];    /* sort structure for stacking list     */ /*960408:ldd*/
        int     nlist;          /* Num. of picks in stacking list       */
        double  tmax;           /* Time range of trav time template     */
        double  ts;             /* Arrival time, stacking phase         */
        double  t1, t2;         /* Stacking time overlap range          */
        int     it;             /* Travel time iterator                 */
        int     ix, iy, iz;     /* Spatial iterators                    */
        int     i, i2, i3;      /* Common iterator                      */
        int     j;              /* Common iterator                      */
        int     is;             /* Stack index                          */
        float   weight;         /* Phase weight (wt dependent)          */
        int     reject;         /* used to reject a pick if higher weight one exists */
        char    txt[80];
        int     l_duplicate;    /* counter for dup pick checks */
        int     islist[MAXLST];
        int     nBnd = 1;

        double  mLikely;        /* Maximum likelyhood value             */  /* SHEEN */
        double  fLikely;        /* Maximum likelyhood value             */  /* SHEEN */
        double  prob;           /* Probablity of a grid point           */  /* SHEEN */
        int     step;
        MLres   mres;

#define MAXHIT 100
        int mhit;               /* Maximum hits in stack                */

        int npix,nused;
        static int ipix[MAXLST]; /* changed to static by WMK            */
        static double tres[MAXLST],tmed,orgt,tmpres[MAXLST];

/* Initialization */
        if(nWt < 1) {
                nWt = 1;
                Wt[0].wt = '0';
                Wt[0].fwt = 1.;
        }
        if (sae != 0) {
                if (time(NULL) > sae) {
                    exit(2);
		}
        }

/* Stacking phase, stack is constructed on the trav times of this phs   */
        ipick = lpick % mPick;
        site = pPick[ipick].site;
        weight = pick_weight( pPick[ipick].wt, pPick[ipick].instid );
        if ( weight < 1.e-4 ) return 0;
        list[0].id  = lpick;
        list[0].t   = 0.;
        list[0].x   = X( Site[site].lon );
        list[0].y   = Y( Site[site].lat );
        list[0].use = 1;
        list[0].horiz = 0;
        list[0].wt  = weight;
        srt[0].t     = pPick[ipick].t;  /*960408:ldd*/
        srt[0].ilist = 0;               /*960408:ldd*/
        if (stack_horizontals != 1 && is_component_horizontal(Site[site].comp[2],Site[site].net)) {
    		if (grid_debug)
                	logit("t", "grid_joint: not using horizontal component to initiate a stack\n");
            list[0].horiz = 1;
            list[0].use = 0;
            goto nada; /* changed in v1.0.5 to just not initiate on a horizontal */
        }

     /* Decide on range of picks to consider in stack */ 
        l2 = lpick + mJoint/2;
        if(l2 > lPick)
                l2 = lPick;
        l1 = l2 - mJoint;
        if(l1 < 0)
                l1 = 0;
        tmax = get_trtime( (nDist-1.)*2./3., nZ-1 );
        ts = pPick[ipick].t;
        t1 = ts - tmax;
        t2 = ts + tmax;

     /* Look thru picks in range; stash info for stackable picks */
        nlist = 1;
        date20(srt[0].t, txt);
        logit("", "########################## l1: %ld, l2: %ld  %s.%s.%s  %s %lf\n",
                            l1,l2, Site[site].name, Site[site].comp, Site[site].net, txt, tmax);
        for(l=l2-1; l>=l1; l--) {   /*go thru pick list backwards*/
                i = l % mPick;
                site_index = pPick[i].site;
//        date20(pPick[i].t, txt);
//        logit("","     GRID_JOINT: %3ld %3d %s.%s.%s  %s\n",l,nlist,Site[site_index].name, Site[site_index].comp, Site[site_index].net, txt);
                if(nlist >= MAXLST)             break;
                if(l == lpick)                  continue;
                if(pPick[i].quake)              continue;
                if(pPick[i].t < t1)             continue;
                if(pPick[i].t > t2)             continue;

                if (stack_horizontals != 1 && is_component_horizontal(Site[site_index].comp[2],Site[site_index].net)) {
                     /* stacking was only designed for P arrivals, so don't use horizontals */
		     if (grid_debug)
                         logit("t", "grid_joint: not using pick on horizontal %s.%s.%s.%s from stashed pick info for stacking against\n",
				Site[site_index].name, Site[site_index].comp, Site[site_index].net, Site[site_index].loc);
		     continue;
        	} 

                /* the following are new checks to deal with multiple picks from the same station */
                if (site == site_index) {
                     /* don't add a pick from the same SCNL to the stackable list */
		     if (grid_debug)
                       logit("t", "grid_joint: not testing pick %s.%s.%s.%s wt(%c) for stacking against same SCNL %s.%s\n",
				Site[site_index].name, Site[site_index].comp, Site[site_index].net, Site[site_index].loc,
                                pPick[i].wt, Site[site].net, Site[site].name);
                     continue;
                }
                /* remember, sites are SCNL's so check S and N and L of pick against initiating phase, 
                   this is the case of multiple Z's on different components*/
                if (is_same_SNL(site_index, site)) {
		     if (grid_debug)
                       logit("t", "grid_joint: not testing pick %s.%s.%s.%s wt(%c) for stacking against same station location %s.%s.%s\n",
				Site[site_index].name, Site[site_index].comp, Site[site_index].net, Site[site_index].loc,
                                pPick[i].wt, Site[site].net, Site[site].name, Site[site].loc);
                     continue;
                }
                /* if we reach here, we have a viable stackable pick (not a horizontal, comp not from same station as initiating pick */

                /* now test the stackable pick against any others from this station, that are not horizontals! */
                reject = 0;   /* set to 1 if this pick should be rejected because there is a better one from this sta in the list */
                for (l_internal=l2-1; l_internal>=l1; l_internal--) {
                     if (l_internal == l) continue; 
                     j = l_internal % mPick;
                     if(pPick[j].quake) continue;	 /* already connected to another quake, don't stack against it */
                     if(pPick[j].t < t1) continue;	/* pick outside time window */
                     if(pPick[j].t > t2) continue;	/* pick outside time window */
                     if (stack_horizontals != 1 && is_component_horizontal(Site[pPick[j].site].comp[2],Site[j].net)) continue;
                     /* if it is a horizontal skip this check */
                     /* okay, this is a viable pick for inclusion, but is it weighted higher then the current one being tested, "l"?? */
                     /* test only if this is same SCNL but diff time pick <or> same SNL but diff time pick */
                     if (pPick[j].site == site_index || is_same_SNL(pPick[j].site, site_index)) {
                          /* remember wt here is 0,1,2,3 which is really quality, 0 better than 1 etc */
                          /* is it case 1, higher weight? (note since wt=quality, < is used below) */
                          if (pPick[j].wt < pPick[i].wt) reject = j;
                          /* is it case 2, equal weight? */
                          if (pPick[j].wt == pPick[i].wt && pPick[j].t > pPick[i].t) reject = j; /* choose earlier one */
                          if (pPick[j].wt == pPick[i].wt && pPick[j].t == pPick[i].t) reject = -1; /* just choose one they are equal */
                          if (reject) break;
                     }
                }
                if (reject > 0) {
                     /* weight was higher or equal to another one in the stack for same site */
		     if (grid_debug)
                       logit("t", "(DEBUG: grid_joint: not stacking against pick %s.%s.%s.%s with quality (%c) lower than a better pick (%c) %s.%s at same sta\n",
				Site[site_index].name, Site[site_index].comp, Site[site_index].net, Site[site_index].loc, 
                                pPick[i].wt, pPick[reject].wt, Site[pPick[reject].site].comp, Site[pPick[reject].site].loc);
                     continue;
                } else {
                     // check previously input picks
		     for (l_duplicate = 0; l_duplicate < nlist; l_duplicate++) {
				j = list[l_duplicate].id % mPick;
				if (pPick[j].site == site_index || is_same_SNL(pPick[j].site, site_index)) {
					if (pPick[j].wt == pPick[i].wt && pPick[j].t == pPick[i].t) {
						if (grid_debug)
							logit("t", "(DEBUG: grid_joint: not stacking against pick %s.%s.%s.%s.. already exist pick at same sta %s.%s\n",
								Site[pPick[j].site].name, Site[pPick[j].site].comp, Site[pPick[j].site].net, Site[pPick[j].site].loc,
								Site[pPick[i].site].comp, Site[pPick[i].site].loc);
						break;    /* SHEEN 2021.01.22 */
					}
				}
			}
            if( l_duplicate != nlist ) continue;    /* SHEEN 2021.01.22 */

                }
                if( fabs(pPick[i].t - ts) > 30. ) 
                {
//                    logit("","     GRID_JOINT:  time residual  %.2f (%.2f, %.2f)\n",fabs(pPick[i].t-ts), pPick[i].t, ts);
                    continue;   /* SHEEN */
                }

                list[nlist].id   = l;
                list[nlist].t   =  pPick[i].t - ts;
                list[nlist].x    = X(Site[site_index].lon);
                list[nlist].y    = Y(Site[site_index].lat);
                list[nlist].use  = 1;                          /*960408:ldd*/
                list[nlist].wt   = pick_weight( pPick[i].wt, pPick[i].instid );
                srt[nlist].t     = pPick[i].t;                 /*960408:ldd*/
                srt[nlist].ilist = nlist;                      /*960408:ldd*/
                if( sqrt(pow(list[nlist].x-list[0].x,2.)+pow(list[nlist].y-list[0].y,2.)) > rMaxBind )
                {
                    if(grid_debug) logit( "t", "grid_joint: not using pick: large distance from %s to %s, %f km\n", 
                        Site[site].name, Site[site_index].name,
                        sqrt(pow(list[nlist].x-list[0].x,2.)+pow(list[nlist].y-list[0].y,2.)) );
                    continue;    // ignore a pick at more than 200 km distance
                }
                nlist++;
        }
//        logit("","     GRID_JOINT:  Nlist  %3d\n",nlist);

/* Look for glitches in list of picks to stack;
   set glitch-pick use-flags to 0; block of code added 960408:ldd. */
        if( GlitchMinPk != 0 )
        {
           qsort( srt, nlist, sizeof(SRT), grid_cmpsrt );
           i  = 0;
           i2 = GlitchMinPk - 1;
           reject = 0;
           while( i2 < nlist ) {
              if( (srt[i2].t-srt[i].t) <= GlitchNsec ) {  /* found a glitch! */
                 for(j=i; j<=i2; j++) {           /* flag all picks in  */
                    list[srt[j].ilist].use = 0;   /* glitch "don't use" */
                    reject ++;
                 }
              }
              i++;
              i2++;
           }
           if( list[0].use == 0 && list[0].horiz == 0) {    /* don't try to stack if master */
              logit( "t",              /* pick belongs to a glitch     */
                     "grid_joint, mhit = 0, glitch\n" );
              goto nada;
           } else if ( list[0].horiz == 1) {
              logit( "t",              /* pick belongs to a horizontal     */
                     "grid_joint, mhit = 0, horizontal\n" );
              goto nada;
           }
        }
    // REMOVE glitch
        if( reject > 0 )
        {
            for(i=0;i<nlist;i++)
            {
                if( list[i].use == 0 )
                {
                    for(i2=i;i2<nlist-1;i2++)
                    {
                        list[i2].id   = list[i2+1].id  ;
                        list[i2].t    = list[i2+1].t   ;
                        list[i2].x    = list[i2+1].x   ;
                        list[i2].y    = list[i2+1].y   ;
                        list[i2].use  = list[i2+1].use ;
                    }
                }
            }
            nlist -= reject;
        }

//        logit("","     GRID_JOINT:  glitch test, %d, reject: %d\n",nlist,reject);
        if( nlist< minPick ) goto nada;

    //**************************************************************
/*  MAXEL */
        ix1 = -1;
        iy1 = -1;
        ix2 = nX;
        iy2 = nY;

        step = 1;
        
        nused = nlist;
//        logit("","     GRID_JOINT:  calcMLE %d\n",nlist);
        mres =  calcMLE(step,list,nlist);
        mLikely = mres.mLikely; 
        xmax    = mres.xmax; 
        ymax    = mres.ymax; 
        izmax   = mres.izmax; 

        // Jackknife
        if( mLikely < minLikely && nlist > minPick )
        {
          MLres   mresT,mresBest;
          fLikely = mLikely;
          tlist[0].id   = list[0].id  ;
          tlist[0].t    = list[0].t   ;
          tlist[0].x    = list[0].x   ;
          tlist[0].y    = list[0].y   ;
          tlist[0].wt   = list[0].wt  ;
          tlist[0].use  = list[0].use ;
          for(is=1;is<nlist;is++)
          {
            for(j=1,i=1; j<is; j++,i++)
            {
                tlist[j].id   = list[i].id  ;
                tlist[j].t    = list[i].t   ;
                tlist[j].x    = list[i].x   ;
                tlist[j].y    = list[i].y   ;
                tlist[j].wt   = list[i].wt  ;
                tlist[j].use  = list[i].use ;
            }
            for(i++; i<nlist; i++,j++)
            {
                tlist[j].id   = list[i].id  ;
                tlist[j].t    = list[i].t   ;
                tlist[j].x    = list[i].x   ;
                tlist[j].y    = list[i].y   ;
                tlist[j].wt   = list[i].wt  ;
                tlist[j].use  = list[i].use ;
            }
            mresT =  calcMLE(step,tlist,nlist-1);
            if( minLikely < mresT.mLikely && fLikely < mresT.mLikely )
            {
                mresBest = mresT;
                i3 = is;
                i2 = pPick[list[is].id % mPick].site;
                date17(pPick[list[is].id % mPick].t, txt);
                logit("","JACKKNIFE: %g  -> %g, removed %d  %-5s  %s  %s  %s\n", fLikely, mresT.mLikely, is,
                                       Site[i2].name, Site[i2].net, Site[i2].comp, txt);
                fLikely = mresT.mLikely;
            }
          }
          if( fLikely > minLikely )
          {
            for(i=0;i<nlist;i++)
            {
                if( i3 == i )
                {
                    for(i2=i;i2<nlist-1;i2++)
                    {
                        list[i2].id   = list[i2+1].id  ;
                        list[i2].t    = list[i2+1].t   ;
                        list[i2].x    = list[i2+1].x   ;
                        list[i2].y    = list[i2+1].y   ;
                        list[i2].use  = list[i2+1].use ;
                    }
                }
            }
            nlist -= 1;
            mres = mresBest;
//            mres =  calcMLE(step,list,nlist);
            mLikely = mresBest.mLikely; 
            xmax    = mresBest.xmax; 
            ymax    = mresBest.ymax; 
            izmax   = mresBest.izmax; 
          }
        }


    //**************************************************************
        for(i=0;i<nlist;i++)
        {
            dtmp = get_trtime( sqrt(pow(list[i].x-xmax,2.)+ 
                                pow(list[i].y-ymax,2.)), izmax);
            tres[i] = dtmp - list[i].t;
            tmpres[i] = tres[i];
        }
        tmed = median(nlist,tmpres);
        orgt = ts - tmed;
        for(i=0;i<nlist;i++) tres[i] -= tmed;

        /* Output for the original MAXEL */
        for(i=0;i<nlist;i++)
        {
            is = pPick[list[i].id % mPick].site;
            date17(pPick[list[i].id % mPick].t, txt);
            logit("","%-5s  %s  %s  1  1.0 1    %2d  %9.3f %6.1f %s %9.4f %9.4f %9.3f\n",
                Site[is].name, Site[is].net, txt,
                i+1,tres[i],sqrt(pow(list[i].x-xmax,2.)+ pow(list[i].y-ymax,2.)),
                Site[is].comp,Site[is].lat,Site[is].lon,list[i].t);
        }
// List phases contributing to stack maximum
        npix = 0;
        for(i=0; i<nlist; i++) {
            if( list[i].use == 0 || list[i].horiz == 1) continue; // skip glitch-picks or horizontal

            // See if the travel-time residual falls in tResBind range
            if(fabs(tres[i]) < tResBind ) {
                ipix[npix++] = i;
            }
            else list[i].use = 0;                               // remove outliers
        }

        /* Summary Output from the MAXEL */
        zmax = grdOrg.z + (double)izmax*dSpaceZ;
        logit( "", "MAXEL: nOrg: %d, nJack: %d, nTthresh: %d, mLikely: %6.4f at %8.4f %9.4f, %5.2f km\n",
                    nused, nlist, npix, mLikely, LAT(ymax), LON(xmax), zmax);

        if( npix < minPick ) 
            goto nada;

// MLE with good picks (nlist was changed after jackknife)
        if( nlist > npix )
        {
          nused = npix;
          for(i=0,j=0; i<nlist; i++)
          {
            if( list[i].use == 1 )
            {
                tlist[j].id   = list[i].id  ;
                tlist[j].t    = list[i].t   ;
                tlist[j].x    = list[i].x   ;
                tlist[j].y    = list[i].y   ;
                tlist[j].wt   = list[i].wt  ;
                tlist[j].use  = list[i].use ;
                j++;
            }
          }

          mres =  calcMLE(step,tlist,nused);
          fLikely = mres.fLikely; 
          mLikely = mres.mLikely; 
          xmax    = mres.xmax; 
          ymax    = mres.ymax; 
          izmax   = mres.izmax; 

          for(i=0;i<nused;i++)
          {
            dtmp = get_trtime( sqrt(pow(tlist[i].x-xmax,2.)+ 
                    pow(tlist[i].y-ymax,2.)), izmax);

            tres[i] = dtmp - tlist[i].t;
            tmpres[i] = tres[i];
          }
          tmed = median(nused,tmpres);
          orgt = ts - tmed;
          for(i=0;i<nused;i++) tres[i] = tres[i] - tmed;

          for(i=0;i<nused;i++)
          {
            is = pPick[tlist[i].id % mPick].site;
            date17(pPick[tlist[i].id % mPick].t, txt);

                logit("","%-5s  %s  %s  1  1.0 1    %2d  %9.3f %6.1f %s %9.4f %9.4f %9.3f *\n",
                    Site[is].name, Site[is].net, txt,
                    i+1,tres[i],sqrt(pow(tlist[i].x-xmax,2.)+ pow(tlist[i].y-ymax,2.)),
                    Site[is].comp,Site[is].lat,Site[is].lon,tlist[i].t);
          }
        /* Summary Output from the MAXEL */
          logit( "", "MAXEL Good, nused: %d, mLikely: %6.4f at %8.4f %9.4f, %5.2f km\n",
                    nused, mLikely, LAT(ymax), LON(xmax), grdOrg.z + ((double)izmax)*dSpaceZ );

    //**************************************************************
// Update List phases contributing to stack maximum
          npix = 0;
          for(i=0; i<nused; i++) {
            for(i2=0;i2<nlist;i2++)
                if( tlist[i].id == list[i2].id ) break;
            // See if the travel-time residual falls in tResBind range
            if(fabs(tres[i]) < tResBind ) {
                ipix[npix++] = i2;
            }
            else  {
                    list[i2].use = 0;                               // remove outliers
                    logit("","MAXEL OUT\n");
            }
          }

        }

    //**************************************************************
//        if( izmax == 0 ) zmax = 1.;
//        else             zmax = ((double)izmax)*dSpaceZ;
        zmax = grdOrg.z + (double)izmax*dSpaceZ;

        prob = 0.8*fLikely;
        mhit = 0;
        for(ix=ix1+1,xx=0.; ix<ix2; ix+=step,xx+=dSpace*step) {
            for(iy=iy1+1,yy=0.; iy<iy2; iy+=step,yy+=dSpace*step) {
                is = IXGRID(ix, iy, izmax);
                if ((is<0) || (is>=dGridMax)) Panic( 3 );
                if( dGrid[is] > prob )
                {
                    mhit++;
                }
            }
        }

    //**************************************************************
        date17(orgt, txt);
        logit("","OT: %s nPicks: %d/%d (%6.4f) at %8.4f %9.4f  %5.1f HIT: %6d\n",
                txt,npix,minPick, mLikely,LAT(ymax),LON(xmax), grdOrg.z + ((double)izmax)*dSpaceZ, mhit);
        // Solution at grid boundary
        nBnd = 1;
        if( mres.cy <= nBnd-1 || mres.cy >= nY-nBnd || mres.cx <= nBnd-1 || mres.cx >= nX-nBnd )
//        if( mres.cy <= nBnd-1 || mres.cy >= nY-nBnd || mres.cx <= nBnd-1 || mres.cx >= nX-nBnd || mres.izmax >= nZ-nBnd )
        {
            logit("","Solution at boundary: ymax: %d / %d, xmax: %d / %d, zmax: %d / %d \n", 
                     mres.cy, nY-1, mres.cx, nX-1, mres.izmax, nZ-1 );
            goto nada;
        }


    //**************************************************************
//        if(npix < minPick || mLikely < minLikely || mhit > 40 )
        if(npix < minPick || mLikely < minLikely )
        {
            logit("","npix: %d / %d, mLikely: %.4f / %.4f, mhit: %d / 40\n",
                    npix,minPick,  mLikely,minLikely, mhit);
            goto nada;
        }

    //**************************************************************
        dthr = 0.;
    /* MEDIAN
        for(i=0,j=0;i<nlist;i++) 
        {
            if( list[i].use == 0 ) continue;
            dtmp = sqrt(pow(list[i].x-xmax,2.)+ pow(list[i].y-ymax,2.));
            tmpres[j++] = dtmp;
        }
        dthr = median(j,tmpres);
    */
    /* MAXIMUM distance */
        for(i=0,j=0;i<nlist;i++) 
        {
            if( list[i].use == 0 ) continue;
            dtmp = sqrt(pow(list[i].x-xmax,2.)+ pow(list[i].y-ymax,2.));
            if( dtmp > dthr ) dthr = dtmp;
        }
        if( dthr > oPDist ) dthr = oPDist;
        /*
        else if( dthr < 30. )
        {
            logit("","MAX DIST (%.2f) is too short (must be greater than 30 km)\n", dthr);
            goto nada;
        }
        */
    //**************************************************************

        for(i=0; i<nSite; i++)
         {
            if( is_component_horizontal(Site[i].comp[2],Site[i].net)) 
                eDist[i] = 1.e5;
            else 
                eDist[i] = sqrt(pow(X(Site[i].lon)-xmax,2.)+ pow(Y(Site[i].lat)-ymax,2.));
            eDid[i] = i;
         }
         qsort_id(eDist, eDid, 0, nSite-1);

         for(i=0,i3=0,it=0; i<nSite; i++)
         {
             i2 = eDid[i];
             if( eDist[i2] < fracDist )
             {
                for(j=0;j<nlist && it<nlist; j++)
                {
                    if(list[j].use == 0 ) continue;            // ignore removed triggers
                    is = pPick[list[j].id % mPick].site;
                    if( i2 == is ) 
                    {
                        it++;
                        break;
                    }
                }
             }
             else break;

             if( it == npix ) break;
         }
         if( it < nTrig )
         {
            logit("", 
                "Triggered stations within threshold:  %d (stations)/%d(threshold_triggered) within %.1f km\n", 
                    it, nTrig, fracDist);
            goto nada;
         }

    //**************************************************************
// Check grid
        if( grid_debug == 2 )
        {
            logit("e","Write grid_info to grd.dat at depth %d \n", izmax);
            FILE *ftmp;
            MLres tmres;

// calcMLE
            hGsearch(0,nX-1,0,nY-1, izmax, step, &tmres, tlist ,nused);

            ftmp = fopen("grd.dat","w");
            iz = izmax;
            for(ix=ix1+1,xx=0.; ix<ix2; ix+=step,xx+=dSpace*step) {
                for(iy=iy1+1,yy=0.; iy<iy2; iy+=step,yy+=dSpace*step) {
                    is = IXGRID(ix, iy, iz);
                    fprintf(ftmp,"%8.4f  %7.4f  %g\n",LON(xx), LAT(yy), log10(dGrid[is]/mLikely));
                }
            }
            fclose(ftmp);
            ftmp = fopen("stn.dat","w");
            for(i=0;i<nlist;i++)
            {
                if( list[i].use == 1 )  fprintf(ftmp,"%8.4f %7.4f 1\n",LON(list[i].x),LAT(list[i].y));
                else                    fprintf(ftmp,"%8.4f %7.4f 0\n",LON(list[i].x),LAT(list[i].y));
            }
            fclose(ftmp);
            ftmp = fopen("loc.dat","w");
            fprintf(ftmp,"%8.4f %7.4f %g\n",LON(xmax), LAT(ymax),mLikely);
            fclose(ftmp);
        }
// Check grid
    //**************************************************************

    //**************************************************************
         // SHEEN 2023.05.10
         if( OPcheck ) 
         {
           nOPsite = 0;
           for( i=0; i<nSite ; i++)
           {
                    if( is_component_horizontal(Site[i].comp[2],Site[i].net)) 
                            continue;
                   if( fabs(LPTime[i] - pPick[ list[0].id % mPick ].t ) > 3600.*OPhour ) 
                           continue;
                    OPsite[nOPsite].lat = Site[i].lat;
                    OPsite[nOPsite].lon = Site[i].lon;
                    OPsite[nOPsite].operating = 1;
                    OPsite[nOPsite].si = i;
                    nOPsite++;
           }
           
         // SHEEN 2021.01.26
           it = i3 = 0;
           l_duplicate = 0; 
           int is2;
           for(i=0; i<nOPsite; i++)   // loop for in-operation stations
           {
             if( OPsite[i].operating == 0 ) continue;
             dtmp = sqrt(pow(X(OPsite[i].lon)-xmax,2.)+ pow(Y(OPsite[i].lat)-ymax,2.));
             is2 = OPsite[i].si;

             if( dtmp < dthr )     // Check distance threshold
             {
                for(j=0;j<nlist;j++)    // loop for valid pick list
                {
                    if(list[j].use == 0 ) continue;            // ignore removed triggers

                    is = pPick[list[j].id % mPick].site;
                    if( is == is2 || is_same_SCNL(is, is2) ) // in-op was triggered
                    {
                        it++;
                          logit("","ASSOCIATED:  %-5s  %s  %s   %6.1f / %6.1f  %9.4f %9.4f \n",
                            Site[is].name, Site[is].net, Site[is].comp,   
                            dtmp, dthr, Site[is].lat,Site[is].lon);
                        break;     
                    }
                    else if( is_same_SNL(is, is2) ) // Same location
                        break;     

/*
                    if( check_pick_list(is, Site[is2].name, Site[is2].net, Site[is2].loc) > 0 )
                    {
                        it++;
                        if( grid_debug )
                        {
                          logit("","NOT ASSOCIATED: %-5s  %s  %6.1f   %9.4f %9.4f  <-> %-5s  %s  %6.1f   %9.4f %9.4f\n",
                            Site[is].name, Site[is].net, 
                            dtmp, Site[is].lat,Site[is].lon,
                            Site[is2].name, Site[is2].net, 
                            dtmp, Site[is2].lat,Site[is2].lon);
                        }
                        break;
                    }
*/
                }
                if( j==nlist ) 
                {
                    is = OPsite[i].si;
                    for(j=0; j<i3; j++)
                    {
                        if( is_same_SNL(is, islist[j]) ) break;
                    }
                    if( j == i3 )
                    {
                        logit("","NOT triggered: %-5s  %s  %6.1f  (%6.1f) %9.4f %9.4f\n",
                            Site[is].name, Site[is].net, 
                            dtmp, dthr, Site[is].lat,Site[is].lon);
                        islist[i3++] = is;
                    }
                }
              }
           }
           
//           logit("", 
//                    "DMedi: Not-triggered: %d (in Opr.), triggered: %d (not Assoc.),  used : %d, stations within %.1f km\n", 
//                     i3, it, npix, dthr);
//           if( i3 >= npix*oPRate || it >= npix*oPRate ) 
//           if( i3 > npix*oPRate || it > npix*oPRate ) 
//           {
//              logit("","CheckOP  %d greater than %.1f (picks*OprRate),  %d  %.1f\n",i3, npix*oPRate, it, npix*oPRate);
//              goto nada;
//           }
           if ( (float)it/(it+i3) - oPRate < 1.e-5 )
           {
              logit(""," %d/%d (TRIGGERED/All_IN-OP) within %.1f km (%.1f) is SMALLER than threshold (%.1f) \n",
                         it, it+i3, dthr, (float)it/(it+i3), oPRate);
              goto nada;
           }
           logit( "", "Go further, npix: %d, not-triggered: %d, triggered: %d (%.1f > %.1f) \n", 
                         npix, i3, it, (float)it/(it+i3), oPRate);
        }
    //**************************************************************

        i = lQuake % mQuake;
        pQuake[i].t    = orgt;
        pQuake[i].lat  = LAT(ymax);
        pQuake[i].lon  = LON(xmax);
        pQuake[i].z    = zmax;
        pQuake[i].rms  = 0.0;
        pQuake[i].dmin = 0.0;
        pQuake[i].ravg = 0.0;
        pQuake[i].gap  = 0.0;
        pQuake[i].npix = npix;
        pQuake[i].nmod = 0;
        pQuake[i].assessed   = 0;
        pQuake[i].lpickfirst = 0;  /* filled in by bind_locate */
        date20(pQuake[i].t, txt);
        logit( "", "  starting location %s %.4f %.4f %.1f\n",
                txt, pQuake[i].lat, pQuake[i].lon, pQuake[i].z);
        logit( "", "grid_joint : npix = %d\n", npix);
        if (is_quake_simultaneous(i)) {
                pQuake[i].npix = 0; /* suggested fix by JinKooLee 5/29/2018 */
                goto nada;      /* terminate nucleation, too close to other quake */
        }
        for(j=0; j<npix; j++) {
                ixpick = ipix[j];
                l = list[ixpick].id % mPick;
                pPick[l].quake = lQuake;  /* assoc w/quake */
                pPick[l].phase = 0;       /* label as P    */ /*bugfix 960408:ldd*/
                bndr_link(lQuake, list[ixpick].id);
        }
        bndr_quake2k(lQuake);
        lQuake++;
        return 1;

nada:
        return 0;
}

/****************************************************************
 *  grid_cmpsrt() compare 2 arrival times in a stacking list.   *
 *                Function added 960408:ldd                     *
 ****************************************************************/
int grid_cmpsrt( const void *l1, const void *l2 )
{
        SRT *lst1;
        SRT *lst2;

        lst1 = (SRT *) l1;
        lst2 = (SRT *) l2;
        if(lst1->t < lst2->t)   return -1;
        if(lst1->t > lst2->t)   return  1;
        return 0;
}

/****************************************************************
 * grid_free()  Free previously malloc'd memory                 *
 ****************************************************************/
void grid_free( void )
{
        free( (void *)dGrid );
        free( (void *)dTrav );
        return;
}
/****************************************************************/
/* are the Site's Station, Network, and Location codes the same */
/* Return 1 if true, 0 if false */
/****************************************************************/
int is_same_SNL(int i1, int i2) {

int loc_is_the_same;
         if (ignore_loc_code_in_same_station_decision) {
           loc_is_the_same = 1; /* ignore the location code check */
         } else {
           loc_is_the_same = (strcmp(Site[i1].loc, Site[i2].loc) == 0); /* actually check loc code */
         }   
         if (strcmp(Site[i1].name, Site[i2].name) == 0 &&
           strcmp(Site[i1].net, Site[i2].net) == 0 &&
           loc_is_the_same) {
           return 1;
         }
         return 0;
}

int is_same_SCNL(int i1, int i2) {
int loc_is_the_same;
         if (ignore_loc_code_in_same_station_decision) {
           loc_is_the_same = 1; /* ignore the location code check */
         } else {
           loc_is_the_same = (strcmp(Site[i1].loc, Site[i2].loc) == 0); /* actually check loc code */
         }   
         if (strcmp(Site[i1].name, Site[i2].name) == 0 &&
           strcmp(Site[i1].net, Site[i2].net) == 0 &&
           strcmp(Site[i1].comp, Site[i2].comp) == 0 &&
           loc_is_the_same) {
           return 1;
         }
         return 0;
}

/* ****************
  is_quake_simultaneous(i) - i is index of quake to test
                returns n which is index of quake that meets test of nearest
                or returns 0 - no nearest quake or test not activated
***************** */
int is_quake_simultaneous(int i) {
int j;
        if (nearest_quake_dist != 0.0) {
                /* check to make sure no nearest neighbor quakes */
                for(j=0;j<mQuake;j++) {
                        /* first check if within time tolerance of other quakes */
                        if (i != j && pQuake[j].npix !=0 && fabs(pQuake[j].t - pQuake[i].t)<nearest_quake_time) {
                                /* then compute distance */
                                double dist, azm;
                                geo_to_km(pQuake[i].lat, pQuake[i].lon, pQuake[j].lat, pQuake[j].lon, &dist, &azm);
                                /* debug */
                                logit("t", "Debug: simultaneous event, time-diff %5.2fs and dist %6.2f km\n",
                                                fabs(pQuake[j].t - pQuake[i].t), dist );
                                if (dist < nearest_quake_dist) {
                                        logit("t", "killing event, too close in time %5.2f < tolerance: %5.2fs and space to another by %6.2f km < tolerance %6.2f km\n",
                                                fabs(pQuake[j].t - pQuake[i].t),nearest_quake_time, dist, nearest_quake_dist);
                                         return(j);
                                }
                        }
                }
        }
        return(0);
}
/****************************************************************
 *  qsort_id() Sort list and save index of list
 ****************************************************************/
void qsort_id(double *a, int *id,int l,int r)
{
   double axis;
   int left, right,tmp;
   if(l<r)
   {
        left = l-1;
        right = r;
        axis = a[id[r]];
        while (left < right)
        {
           while (a[id[++left]] < axis);
           while (a[id[--right]] > axis) if(right==0) break;

           if (left >= right ) break;
           tmp = id[left];
           id[left] = id[right];
           id[right] = tmp;
        }
        tmp = id[left];
        id[left] = id[r];
        id[r] = tmp;
        qsort_id(a,id,l,left-1);
        qsort_id(a,id,left+1,r);
   }
}

int check_pick_list(int is, char *stn, char *ntw, char *loc)
{
    int loc_is_the_same;
    if (ignore_loc_code_in_same_station_decision) 
    {
        loc_is_the_same = 1; /* ignore the location code check */
    } 
    else 
    {
        loc_is_the_same = (strcmp(Site[is].loc, loc) == 0); /* actually check loc code */
    }   
    if (strcmp(Site[is].name, stn) == 0 &&
        strcmp(Site[is].net, ntw) == 0 &&
        loc_is_the_same) 
    {
        return 1;
    }
    return 0;
}

QUAKE testMLE(int qid, int npix, int *IPLIST)                 /* SHEEN */
{
    int i,ip,site,it;
    double tres[MAXLST],tmed,orgt,trt;
    PLIST   mlist[MAXLST];
    QUAKE tQuake;
    MLres mres;
    char txt[5000];


//    logit("t","testMLE: %8.4f %8.4f %.2f  %d\n",
//        pQuake[qid].lat, pQuake[qid].lon, pQuake[qid].z, npix);
    for(i=0;i<npix;i++)
    {
        ip = IPLIST[i];
        site = pPick[ip].site;

        mlist[i].t   =  pPick[ip].t -  pPick[ IPLIST[0] ].t;
        mlist[i].x   = X( Site[site].lon );
        mlist[i].y   = Y( Site[site].lat );
        mlist[i].wt  = pick_weight( pPick[ip].wt, pPick[ip].instid );
    }

    for(i=0;i<npix;i++)
    {
        ip = IPLIST[i];
        date20(pPick[ip].t, txt);
        site = pPick[ip].site;
//        logit("t","testMLE: %s.%s.%s  %8.4f %8.4f %8.2f %s %d\n", Site[site].name, Site[site].comp, Site[site].net,
//            Site[site].lon, Site[site].lat, mlist[i].t,txt, ip);
    }

    mres = calcMLE(1, mlist, npix);               

    for(i=0;i<npix;i++)
    {
        trt = get_trtime( sqrt(pow(mlist[i].x-mres.xmax,2.)+ 
                    pow(mlist[i].y-mres.ymax,2.)), mres.izmax);
        tres[i] = trt - mlist[i].t;
    }
    tmed = median(npix,tres);
    orgt = mlist[0].t - tmed;

    tQuake.t    = orgt;
    tQuake.lat  = LAT(mres.ymax);
    tQuake.lon  = LON(mres.xmax);
    tQuake.z    = ((double)mres.izmax+0.5)*dSpace;

 
    logit("t","testMLE: %8.4f %8.4f %.2f    %8.4f  %8.4f  %2.f\n",
        pQuake[qid].lat, pQuake[qid].lon, pQuake[qid].z,
        tQuake.lat, tQuake.lon, tQuake.z); 


    return tQuake;
}
 
