#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "earthworm.h"
/*
#include "smoothing.h"
*/

void konno_ohmachi_c(double *spec, double *freqs, int npoints, double *ko_freqs,
                     double *ko_smooth, int nko, double bandwidth);

float get_peak_freq(double *spec, double dt, int nfft,
                       double bandwidth) {

  int i, ii, nfreq, nkofreqs;
  double df, dre, dim;
  double max_val, peakf, dlogf, logf;
  double f, f1, f2, taper;
  double *amp, *freq, *ko_smooth, *ko_freqs;

  //printf("MTH: Enter get_peak_freq nfft=%d dt=%f bandwidth=%f\n", nfft, dt, bandwidth);

  nfreq = nfft/2;

  // allocate amp, freq
  if ( (amp = (double *)calloc( nfreq+1, sizeof(double))) == (double *)NULL)
  {
    logit("et", "Configure: out of memory for amp array\n");
    return -1;
  }
  if ( (freq = (double *)calloc( nfreq+1, sizeof(double))) == (double *)NULL)
  {
    logit("et", "Configure: out of memory for freq array\n");
    return -1;
  }

  // rotate spec to amp 0,..., nfreq-1
  amp[0] = 0;
  amp[nfreq] = 0;            // Set Nyquist value to 0
  df = 1.0 / (nfft * dt);
  freq[0] = 0;
  freq[nfreq] = nfft/2 * df; // Nyquist freq
  f1 = 0.;
  f2 = 0.33;
  for (i = 1; i < nfreq; i++)
  {
    ii = i+i;
    dre = spec[ii];   /* Real part of transformed data */
    dim = spec[ii+1]; /* Imaginary part of transformed data */
    //printf("%4d: %12.5e %12.5e\n", i, dre, dim);
    amp[i] = sqrt(dre*dre + dim*dim);
    freq[i] = i * df;
    if (freq[i] <= f2) { // Taper f's near DC 
	f = freq[i];
        taper = exp(-(f2-f)/(f-f1));
	amp[i] *= taper;
    }
  }


  // allocate ko_smooth, kofreqs
  nkofreqs = 301;
  if ( (ko_smooth = (double *)calloc( nkofreqs, sizeof(double))) == (double *)NULL)
  {
    logit("et", "Configure: out of memory for ko_smooth array\n");
    return -1;
  }
  if ( (ko_freqs = (double *)calloc( nkofreqs, sizeof(double))) == (double *)NULL)
  {
    logit("et", "Configure: out of memory for ko_freqs array\n");
    return -1;
  }

  // make log-spaced ko_freqs
  dlogf = (log10(freq[nfreq]) - log10(freq[1])) / (nkofreqs-1);
  //printf("MTH: dlogf=%f\n", dlogf);
  for (i=0; i<nkofreqs; i++){
    logf = log10(freq[1]) + i*dlogf;
    ko_freqs[i] = pow(10, logf);
    //printf("i:%d logf:%f f:%f\n", i, logf, ko_freqs[i]);
  }
  //printf("freq[1]=%f freq[%d]=%f\n", freq[1], nfreq, freq[nfreq]);
  //printf("ko_freqs[0]=%f ko_freqs[1]=%f ko_freqs[%d]=%f\n", ko_freqs[0], ko_freqs[1], nkofreqs-1, ko_freqs[nkofreqs-1]);

  //printf("MTH: get_peak_period nfreq=%d\n", nfreq);

  // call konno_ohmachi_c to fill ko_smooth
  konno_ohmachi_c(amp, freq, nfreq,
                  ko_freqs, ko_smooth, nkofreqs,
                  bandwidth);

  // find peak of ko_smooth and return
  max_val = 0.;
  peakf = 0.;
  for (i=0; i<nkofreqs; i++){
    if (ko_smooth[i] > max_val){
      max_val = ko_smooth[i];
      peakf = ko_freqs[i];
    }
  }
  //printf("max_val=%12.6e at f:%f\n", max_val, peakf);


f_exit:
  // Deallocate amp, freq, ko_freqs, ko_smooth
  free(amp);
  amp = (double *)NULL;
  free(freq);
  freq = (double *)NULL;
  free(ko_freqs);
  ko_freqs = (double *)NULL;
  free(ko_smooth);
  ko_smooth = (double *)NULL;

  return peakf;
}

/*
 * This code implements Konno-Ohmachi spectral smoothing as defined in their
 * paper:
 * Konno, K. and Ohmachi, T., 1998. Ground-motion characteristics estimated
 * from spectral ratio between horizontal and vertical components of
 * microtremor. Bulletin of the Seismological Society of America, 88(1),
 * pp.228-241.
 * The function below is partially based on a Python version by Albert
 * Kottke and found here:
 * https://github.com/arkottke/notebooks/blob/master/effective_amp_spectrum.ipynb
 * It was rewritten and optimized in C by Bruce Worden.
 */
void konno_ohmachi_c(double *spec, double *freqs, int npoints,
                     double *ko_freqs, double *ko_smooth, int nko,
                     double bandwidth) {
    int i, j;
    double window_total, total, x, fc, freq, frat, window;
    double max_ratio = pow(10.0, (3.0 / bandwidth));
    double min_ratio = 1.0 / max_ratio;

    /*
    printf("Enter konno_ohmachi_c npoints:%d nko:%d bandwidth:%f\n", npoints, nko, bandwidth);
    for(j = 0; j < npoints; j++) {
	    printf("%12.6e\n", spec[j]);
    }
    exit(-1);
    */

    for(i = 0; i < nko; i++) {
        fc = ko_freqs[i];
        if (fc < 1e-6) {
            ko_smooth[i] = 0;
            //ko_smooth[i] = NAN;
            continue;
        }
        total = 0;
        window_total = 0;
        for(j = 0; j < npoints; j++) {
            freq = freqs[j];
            frat = freq / fc;
            if (freq < 1e-6 ||
                frat > max_ratio || frat < min_ratio) {
                continue;
            } else if (fabs(freq - fc) < 1e-6) {
                window = 1.0;
            } else {
                x = bandwidth * log10(frat);
                window = sin(x) / x;
                window *= window;
                window *= window;
            }
            total += window * spec[j];
            window_total += window;
        }
        if (window_total > 0) {
            ko_smooth[i] = total / window_total;
        } else {
            ko_smooth[i] = 0;
            //ko_smooth[i] = NAN;
        }
    }
    /*
    for(j = 0; j < nko; j++) {
       printf("%12.6e\n", ko_smooth[j]);
    }
    exit(-1);
    */
    return;
}
